/*BEGIN_LEGAL 
  Intel Open Source License 

  Copyright (c) 2002-2007 Intel Corporation 
  All rights reserved. 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */
/*
 *  This file contains an ISA-portable PIN tool for tracing memory accesses.
 */

/*
 * Modified on Sep 29, 2008
 * Author: Sandeep Budanur Ramanna
 * E-mail: sbudanu@ncsu.edu
 * North Carolina State University
 */

/* if following two stackwalks are undefined, Ver0(Naive) Stackwalk is used */
#define WISC_STACKWALK_ENABLED 1
//#define LIBUNWIND_STACKWALK_ENABLED 1
#define OPTIMIZED_STACKWALK_ENABLED 1

#include <stdio.h>
#include "pin.H"
#include "instlib.H"
#include <pthread.h>
#include <stdlib.h>
#include <setjmp.h>
#include "format.h"
#include <mpi.h>
#include <time.h>
#include <sys/time.h>
#ifdef WISC_STACKWALK_ENABLED 
#include "walker.h"
#include "frame.h"
#elif LIBUNWIND_STACKWALK_ENABLED
//#include <libunwind.h>
#endif

//#define STDOUT
//#define PIPE 1
#define LIMIT 1000
#define MAX_REFS 1000000 // we want to limit the size of the log files so we will limit trace files to 2M lines
#define MAX_THREADS 2048
#define TRACE_DIR "./TRACE_DIR"
#define SKIP_DEPTH    0 //4 /* not used */
#define TRACE_DEPTH 10 //1000 /* maximum length of stack sig */
#define PRSD_PIPE "/home/sbudanu/prsd_pipe"  //"tmp.trace" //"/home/sbudanu/prsd_pipe"
#define SP_BP_DIFF 100000
#define BIN_MARGIN (unsigned long)(0x0000000000FFFFFFUL) //16MB
#define ABS(a) (a>=0?a:-a)
#define RANGE(x) (unsigned long)((unsigned long)x & (unsigned long)(0xFFFFFFFFFF000000UL))
#define TRACE_TAG 0xBEADBEAD


#if 0
#define LOAD_OUTPUT_FORMAT "ip:%20p ebp:%20p sp:%20p sign: %20p tid:%4u l addr:%12p  %u\n", \
	ip,     ebp,     sp,     (void*)sign,      thread_id, addr,     sincelast
#define STORE_OUTPUT_FORMAT "ip:%20p ebp:%20p sp:%20p sign: %20p tid:%4u s addr:%12p  %u\n", \
	ip,     ebp,     sp,     (void*)sign,      thread_id, addr,     sincelast
#endif

#ifdef WISC_STACKWALK_ENABLED
using namespace std;
using namespace Dyninst;
using namespace Stackwalker;
#endif

FILE* logfptr = NULL;
long prev_ip[MAX_THREADS];
//FILE * trace;
struct timeval start_time, end_time;
struct timeval pgm_start_time, pgm_end_time;
struct timeval swlk_start_time, swlk_end_time;
time_t elapsed_time;
time_t swalk_time = 0;

//application's main routine's address is stored 
AFUNPTR mainFunAddr = 0;

INSTLIB::ICOUNT icount;
UINT32 icount_lastmem;
UINT32 sincelast;

jmp_buf jmp_buffer;
UINT32 memrefs; // number of memory accesses so far
UINT32 sigcount = 0;

#ifdef MULTI_FILE //Sandeep: need not save to different files. Just print to stdout and pipe to prsd_comp binary
FILE * fptr[MAX_THREADS];
char fileName[256];
#endif

enum
{
	INS_TYPE_LOAD = 0x11223344AABBCCDD,
	INS_TYPE_STORE = 0x556677889900EEFF
};

int main(int argc, char *argv[]);

#include "myqueue.cpp"
LIST<long> signList;
long xor_sign;

INSTLIB::CONTROL control;
bool doinst;
bool stackwalk_required = true;
PIN_LOCK lock;

#if defined(__x86_64__)

#ifdef WISC_STACKWALK_ENABLED
vector<Frame> stackwalk;
string s;
Walker *walker = NULL;
Frame *frame = NULL;
#endif

#ifdef WISC_STACKWALK_ENABLED
int getSignatureW(MachRegisterVal pc, MachRegisterVal ra, MachRegisterVal fp, MachRegisterVal sp, unsigned long* sign, UINT32 thread_id)
{
	//Walker *walker = Walker::newWalker();
	//Frame *frame = Frame::newFrame(ra, sp, fp, walker);
	//printf("\nWisc Stackwalk called");
	sigcount++;
	gettimeofday(&swlk_start_time, NULL);
	if(!walker)
	{
		walker = Walker::newWalker();
		//printf("\nwalker allocated\n");
	}
	if(!frame)
	{
		frame = Frame::newFrame(0,0,0,walker);
		//printf("\nframe allocated\n");
	}

	if(!walker || !frame)
	{
		printf("\nNULL walker or frame...exiting");
		exit(1);
	}
	frame->setRA(ra);
	frame->setSP(sp);
	frame->setFP(fp);
	walker->walkStackFromFrame(stackwalk, *frame);
	*sign = 0;
	signList.removeAll();

	/* ignoring the last return address value so that the signature across multiple threads remains same */
	for (unsigned i=0; i<(stackwalk.size()-1); i++)
	{
		signList.addHead((long)stackwalk[i].getRA());
		(*sign) ^= (unsigned long)(stackwalk[i].getRA());
	}
	//delete frame;
	//delete walker;
	gettimeofday(&swlk_end_time, NULL);
	elapsed_time = (swlk_end_time.tv_sec * 1000000 + swlk_end_time.tv_usec) - (swlk_start_time.tv_sec * 1000000 + swlk_start_time.tv_usec);
	swalk_time += elapsed_time;	
	return 0;
}
#elif LIBUNWIND_STACKWALK_ENABLED /* LIBUNWIND doesn't work with PIN correctly, results in crash */
#define UNW_LOCAL_ONLY
extern int getSignatureLibunwind(void** _pc, void** _ra, void** _fp, void** _sp, unsigned long* sign);
int getSignatureL(void** _pc, void** _ra, void** _fp, void** _sp, unsigned long* sign)
{
	sigcount++;
	printf("\ngetSigL()");
	//getSignatureLibunwind(_pc, _ra, _fp, _sp, sign);
#if 0
	unw_cursor_t cursor; unw_context_t uc;
	unw_word_t ip, sp, bp;

	unw_getcontext(&uc);
	unw_init_local(&cursor, &uc);
	*sign = 0;
	while (unw_step(&cursor) > 0) 
	{
		unw_get_reg(&cursor, UNW_REG_IP, &ip);
		unw_get_reg(&cursor, UNW_REG_SP, &sp);
		unw_get_reg(&cursor, UNW_TDEP_BP, &bp);
		printf ("ip = %lx, sp = %lx, bp = %lx\n", (long) ip, (long) sp, (long)bp);
		(*sign) ^= (unsigned long)ip;
	}
#endif

	return 0;
}  
#else /* original stackwalk */
__inline__ static void **NextStackFrame(void **old_sp) 
{
	//fprintf(stdout, "\n############9");
	if(!old_sp)
		return NULL;

	void **new_sp = (void **) *old_sp;
	//fprintf(stdout, "\n############8");

	if(!new_sp)
		return NULL;

	//fprintf(stdout, "\n############7");

	//fprintf(stdout, "##new-sp:%p old_sp:%p##\n",(void*)new_sp, (void*)old_sp);

	if (new_sp <= old_sp) 
	{
		//fprintf(stdout, " (11111) ");
		return NULL;
	}
	if ((uintptr_t)new_sp & (sizeof(void *) - 1)) 
	{
		//fprintf(stdout, " (22222) ");
		return NULL;
	}
	if ((uintptr_t)new_sp - (uintptr_t)old_sp > 10000) 
	{
		//fprintf(stdout, " (33333) ");
		return NULL;
	}
	return new_sp;
}

/*
 * perform stackwalk and compute the signature
 */
int getSignature(void** pc, void** sfp, long* sign)
{
	int n = 0;
	void** new_sfp = NULL;
	sigcount++;

	if(!sign || !pc || !sfp)  
		return -1;

	signList.removeAll();
	*sign = 0;
	//signList.addTail(*sign);
	//fprintf(stdout, "\nver0############1 - %p", (void*)sfp); 	

	while (sfp && n < TRACE_DEPTH-1) 
	{
		//fprintf(stdout, "\n###########2");
		if((sfp))
		{	
			if (*(sfp) == (void *)0) //return address is NULL
			{
				//fprintf(stdout, "(1)");
				break;
			}
		}    
		//fprintf(stdout, " RA: %16p ",(long*)*(sfp+1)); 
		//fprintf(stdout, "OLD PC: %p, RA: %p, FP: %p", (void*)pc, (void*)*(sfp+1),(void*)sfp);
		(*sign) ^= (long)(*(sfp+1)); //XOR PC and return addresses to compute unique signature
		//fprintf(stdout, "#Sign: %16p# ",(long*)(*sign)); 
		//signList.addTail((long)(*(sfp+1)));
		signList.addHead((long)(*(sfp+1)));

		if(sfp)
		{		
			new_sfp = NextStackFrame(sfp);
		}	

		if (!new_sfp) 
		{
			//fprintf(stdout, "(2)");
			break;
		}    
		n++;
		sfp = new_sfp;
	}

	//fprintf(stdout, "\n");

	return 0;
}
#endif /* #ifdef WISC_STACKWALK_ENALBED */
#endif /* #if defined(__x86_64__) */

// update lastmem info and write to file (load)
VOID RecordMemRead(VOID * ip, VOID * addr, UINT32 thread_id, VOID *sp, VOID* ebp, VOID* br_taddr, VOID* retAddr, CONTEXT* ctxt)
{
	//long sign = 0;
	char valid = 'N';
	//ADDRINT val = 0;

	/* maximum traces reached */
	if(memrefs > MAX_REFS) 
	{
		//LOG("\nremoving instrumentation in read");
		doinst = false;
		return;
	}

	GetLock(&lock, thread_id+1);

#if 0   //PIN2.7 compilation error 
	if(icount.CountWithREP() == icount_lastmem) 
	{sincelast = 0;}
	else 
	{
		sincelast = icount.CountWithREP() - icount_lastmem - 1;
		icount_lastmem = icount.CountWithREP();
	}
#endif

	if(thread_id < 0)
	{
		cout<<"Invalid Thread ID."<<endl;
		PIN_Detach();
		ReleaseLock(&lock);
		return;
	}

#ifdef WISC_STACKWALK_ENABLED
	{
		if(RANGE(mainFunAddr)== RANGE(ip))
		{	
			if(stackwalk_required)
			{
#ifdef OPTIMIZED_STACKWALK_ENABLED
				stackwalk_required = false;
#endif
				getSignatureW((MachRegisterVal)ip, (MachRegisterVal)retAddr, (MachRegisterVal)ebp, (MachRegisterVal)sp, (unsigned long*)&xor_sign, thread_id);
			}
			//else
			//	printf("\nStackwalk ignored");
			valid = 'V';
		}
	}
#elif LIBUNWIND_STACKWALK_ENABLED
	{
		long* sf_ptr = (long*)ebp; 
		//fprintf(stdout,"gen sf_ptr: %16p\n", sf_ptr);
		if(((unsigned long)sf_ptr >= (unsigned long)sp) &&\
				(((unsigned long)sf_ptr - (unsigned long)sp) <= SP_BP_DIFF) && \
				(RANGE(mainFunAddr)== RANGE(ip))
		  )
		{
			if(stackwalk_required)
			{
#ifdef OPTIMIZED_STACKWALK_ENABLED
				stackwalk_required = false;
#endif
				getSignatureL((void**)ip, (void**)retAddr, (void**)ebp, (void**)sp, (unsigned long*)&xor_sign);
			}
			valid = 'V';
		}
	}
#else
	{
		long* sf_ptr = (long*)ebp; 
		//fprintf(stdout,"gen sf_ptr: %16p\n", sf_ptr);
		if(((unsigned long)sf_ptr >= (unsigned long)sp) &&\
				(((unsigned long)sf_ptr - (unsigned long)sp) <= SP_BP_DIFF) && \
				(RANGE(mainFunAddr)== RANGE(ip))
		  )
		{
			if(stackwalk_required)
			{
#ifdef OPTIMIZED_STACKWALK_ENABLED
				stackwalk_required = false;
#endif
				getSignature((void**)ip, (void**)ebp,  &xor_sign);
			}
			valid = 'V';
		}
	}
#endif
	//fprintf(logfptr, "\n$$$ip: %p %c sfp: %p, sp: %p$$$", (void*)ip, valid, ebp, sp); 
	//fprintf(stdout, "\n#Sign:%20p\n", (void*)sign);
	if(valid == 'V')
	{
		long val = 0;
		xor_sign = 0;
		LIST_ITERATOR<long> *iter = signList.createIterator(); 
		fprintf(logfptr, "\n*%ld*", signList.getSize()+1); 
		fprintf(logfptr, "\n<");
		while(iter->next())
		{
			//fprintf(logfptr,"%p, ", (void*)iter->currentItem()); 
			val = iter->currentItem();
			xor_sign ^= val;
			//write ret values list (signature)
			fprintf(logfptr, " %20p", (void*)val);
		}
		signList.releaseIterator(iter);
		xor_sign = xor_sign ^ (long)ip;
		val = (long)ip;
		if((long)ip == prev_ip[thread_id])
		{
			val = val ^ 'l';
			xor_sign = xor_sign ^ 'l';
		}	
		prev_ip[thread_id] = (long)ip;
		fprintf(logfptr, " %20p", (void*)val);
		//signList.removeAll();
		fprintf(logfptr, ">\n");
		fprintf(logfptr, LOAD_OUTPUT_FORMAT);
		++memrefs;
	}

	if(memrefs > MAX_REFS) 
	{
		cout << "Maximum number of memops reached. Stopping instrumentation... " << endl;
		// if(logfptr)
		//  {fclose(logfptr); logfptr = NULL;}
		//PIN_Detach();
	}

	if(memrefs%LIMIT == 0)
	{
		//printf("\n%d refs. written", memrefs);	
	}

	ReleaseLock(&lock); 
}

// Print a memory write record
VOID RecordMemWrite(VOID * ip, VOID * addr, UINT32 thread_id, VOID* sp, VOID* ebp, VOID* br_taddr, VOID* retAddr, CONTEXT* ctxt)
{
	//long sign = 0;
	char valid = 'N';
	//ADDRINT val = 0;

	/* maximum traces reached */
	if(memrefs > MAX_REFS) 
	{
		//LOG("\nremoving instrumentation in read");
		doinst = false;
		return;
	}

	GetLock(&lock, thread_id+1);
#if 0
	if(icount.CountWithREP() == icount_lastmem) 
	{sincelast = 0;}
	else 
	{
		sincelast = icount.CountWithREP() - icount_lastmem - 1;
		icount_lastmem = icount.CountWithREP();
	}
#endif

	if(thread_id < 0)
	{
		cout<<"Invalid Thread ID."<<endl;
		PIN_Detach();
		ReleaseLock(&lock);
		return;
	}

#ifdef WISC_STACKWALK_ENABLED 
	if(RANGE(mainFunAddr) == RANGE(ip))
	{
		if(stackwalk_required)
		{
#ifdef OPTIMIZED_STACKWALK_ENABLED
			stackwalk_required = false;
#endif
			getSignatureW((MachRegisterVal)ip, (MachRegisterVal)retAddr, (MachRegisterVal)ebp, (MachRegisterVal)sp, (unsigned long*)&xor_sign, thread_id);
		}
		//else
		//	printf("\nStackwalk ignored");
		valid = 'V';             
	}
#elif LIBUNWIND_STACKWALK_ENABLED
	{
		long* sf_ptr = (long*)ebp;
		if((unsigned long)sf_ptr >= (unsigned long)sp &&\
				(((unsigned long)sf_ptr - (unsigned long)sp) <= SP_BP_DIFF) && \
				(RANGE(mainFunAddr) == RANGE(ip))
		  )
		{
			if(stackwalk_required)
			{
#ifdef OPTIMIZED_STACKWALK_ENABLED
				stackwalk_required = false;
#endif
				getSignatureL((void**)ip, (void**)retAddr, (void**)ebp, (void**)sp, (unsigned long*)&xor_sign);
			}
			valid = 'V';             
		}
	}
#else
	{
		long* sf_ptr = (long*)ebp;
		if((unsigned long)sf_ptr >= (unsigned long)sp &&\
				(((unsigned long)sf_ptr - (unsigned long)sp) <= SP_BP_DIFF) && \
				(RANGE(mainFunAddr) == RANGE(ip))
		  )
		{
			if(stackwalk_required)
			{
#ifdef OPTIMIZED_STACKWALK_ENABLED
				stackwalk_required = false;
#endif
				getSignature((void**)ip, (void**)ebp,  &xor_sign);
			}
			valid = 'V';
		}
	}
#endif

	if(valid == 'V')
	{
		long val = 0;
		xor_sign = 0;
		LIST_ITERATOR<long> *iter = signList.createIterator(); 
		fprintf(logfptr, "\n*%ld*", signList.getSize()+1); 
		fprintf(logfptr, "\n<");
		while(iter->next())
		{
			//fprintf(logfptr,"%p, ", (void*)iter->currentItem()); 
			val = iter->currentItem();
			xor_sign ^= val;
			//write ret values list (signature)
			fprintf(logfptr, " %20p", (void*)val);
		}
		signList.releaseIterator(iter);
		xor_sign = xor_sign ^ (long)ip;
		val = (long)ip;
		if((long)ip == prev_ip[thread_id])
		{
			val = val ^ 's';
			xor_sign = xor_sign ^ 's';
		}	
		prev_ip[thread_id] = (long)ip;
		fprintf(logfptr, " %20p", (void*)val);
		//signList.removeAll();
		fprintf(logfptr, ">\n");
		fprintf(logfptr, STORE_OUTPUT_FORMAT);
		++memrefs;
	}


	if(memrefs > MAX_REFS) 
	{
		cout << "Maximum number of memops reached. Stopping instrumentation... " << endl;
		//if(logfptr)
		// {fclose(logfptr); logfptr = NULL;}
		//PIN_Detach();
	}

	if(memrefs%LIMIT == 0)
	{
		//printf("\n%d refs. written", memrefs);	
	}

	ReleaseLock(&lock); 
}


VOID ImageLoad(IMG img, VOID * v)
{
	IMG_TYPE imgType = IMG_Type(img);

	//cout<<"\n"<<"!!!!IMAGE LOAD 1!!!!"<<IMG_Name(img)<<"\n";

	if(imgType ==  IMG_TYPE_STATIC || imgType == IMG_TYPE_SHARED)
	{
		RTN mainRtn = RTN_FindByName(img, "main");
		//cout<<"\n"<<"!!!!IMAGE LOAD 2!!!!\n";
#if 0
		//scan symbol table for main() address 
		for(SYM sym = IMG_RegsymHead(img); SYM_Valid(sym); sym = SYM_Next(sym))
		{    
			//cout<<"\n&&&"<<"Name: "<<SYM_Name(sym);

			if( strcmp(SYM_Name(sym).c_str(), "main") == 0) 
			{
				mainFunAddr = (AFUNPTR)SYM_Value(sym);  
				cout<<"\n"<<"!!!!MAIN ADDR FOUND!!!!\n"; 
			}

			fprintf(stdout, " Addr: %20p&&&\n",(void*)SYM_Value(sym));
		}
#endif      
		if (RTN_Valid(mainRtn))//save main()'s address to compare the PC values later
		{
			mainFunAddr = RTN_Funptr(mainRtn);
			//cout<<"\n"<<"!!!!MAIN ADDR FOUND *!!!!\n";
			fprintf(stdout, "\nmain()-addr: %20p", mainFunAddr);
		}
	}
}

VOID funcall(RTN rtn)
{
	//cout<<RTN_Name(rtn)<<" called\n";
	stackwalk_required = true;
}

VOID funret(RTN rtn)
{
	//cout<<RTN_Name(rtn)<<" returned\n";
	stackwalk_required = true;
}

// Pin calls this function every time a new rtn is executed
VOID Routine(RTN rtn, VOID *v)
{   
#if 0
	ADDRINT fun_addr = RTN_Address(rtn);
	string fun_name = RTN_Name(rtn);
	cout<<"\nfunction name:"<<fun_name;
	fprintf(stdout, "fun addr: %20p", (void*)fun_addr);
#endif	

	RTN_Open(rtn);
	RTN_InsertCall(rtn, IPOINT_BEFORE, (AFUNPTR)funcall, IARG_PTR, rtn, IARG_END);
	RTN_InsertCall(rtn, IPOINT_AFTER, (AFUNPTR)funret, IARG_PTR, rtn, IARG_END);
	RTN_Close(rtn);

}

VOID ThreadBegin(UINT32 threadid, VOID * sp, int flags, VOID *v)
{
	GetLock(&lock, threadid+1);
	fprintf(stdout, "thread begin %d sp %p flags %x\n",threadid, sp, flags);
	ReleaseLock(&lock);
}

VOID ThreadEnd(UINT32 threadid, INT32 code, VOID *v)
{
	GetLock(&lock, threadid+1);
	fprintf(stdout, "thread end %d code %d\n",threadid, code);
	ReleaseLock(&lock);
}

VOID ContextChangeCb(THREADID threadid, CONTEXT_CHANGE_REASON reason, const CONTEXT *from, CONTEXT *to, INT32 info, VOID *v)
{
	GetLock(&lock, threadid+1);
	fprintf(stdout, "thread context switch id:%d reason:%d ctx_from:%p ctx_to:%p\n",threadid, reason, from, to);
	ReleaseLock(&lock);
}

// Is called for every instruction and instruments reads and writes
VOID Instruction(INS ins, VOID *v)
{
	if(doinst) {

		// instruments loads using a predicated call, i.e.
		// the call happens iff the load will be actually executed
		// (this does not matter for ia32 but arm and ipf have predicated instructions)
		if (INS_IsMemoryRead(ins))
		{
			INS_InsertPredicatedCall(
					ins, IPOINT_BEFORE, (AFUNPTR)RecordMemRead,
					IARG_INST_PTR,
					IARG_MEMORYREAD_EA,
					IARG_THREAD_ID,
					IARG_REG_VALUE, REG_STACK_PTR,
					IARG_REG_VALUE, REG_RBP,
					IARG_BRANCH_TARGET_ADDR,
					IARG_RETURN_IP,
					IARG_CONTEXT,
					IARG_END);
		}

		// instruments stores using a predicated call, i.e.
		// the call happens iff the store will be actually executed
		if (INS_IsMemoryWrite(ins))
		{
			INS_InsertPredicatedCall(
					ins, IPOINT_BEFORE, (AFUNPTR)RecordMemWrite,
					IARG_INST_PTR,
					IARG_MEMORYWRITE_EA,
					IARG_THREAD_ID, 
					IARG_REG_VALUE, REG_STACK_PTR,
					IARG_REG_VALUE, REG_RBP,
					IARG_BRANCH_TARGET_ADDR,  
					IARG_RETURN_IP,
					IARG_CONTEXT,
					IARG_END);
		}
	}
}

VOID Handler(INSTLIB::CONTROL_EVENT ev, VOID * v, CONTEXT * ctxt, VOID * ip, THREADID tid)
{
	switch(ev)
	{
		case INSTLIB::CONTROL_START:
			cout << "\nStarting instrumentation..." << tid <<endl;
			doinst = true;
			gettimeofday(&start_time, NULL);
#if 0
			icount_lastmem = icount.CountWithREP();
#endif
			break;

		case INSTLIB::CONTROL_STOP:
			gettimeofday(&end_time, NULL);
			elapsed_time = (end_time.tv_sec * 1000000 + end_time.tv_usec) - (start_time.tv_sec * 1000000 + start_time.tv_usec);
			printf("\nStart inst to Stop inst time = %ld usec\n", elapsed_time);
			cout << "\nStopping instrumentation..." << tid <<endl;
			doinst = false;
			break;

		default:
			ASSERTX(false);
			break;
	}
}

VOID Fini(INT32 code, VOID *v)
{
	//fclose(trace);
	printf("......Fini()..........\n");

	if(logfptr)
	{
		fflush(logfptr);
		printf("\nclosing file");
		fclose(logfptr);
		logfptr = NULL;
	}
	printf("\nTotal stackwalk time: %ld usec\n", swalk_time);
	printf("\ngetSign() function count: %d\n", sigcount);
	printf("\ntime per getSign(): %.4f usec\n", swalk_time * 1.0/sigcount);
	gettimeofday(&pgm_end_time, NULL);
	elapsed_time = (pgm_end_time.tv_sec * 1000000 + pgm_end_time.tv_usec) - (pgm_start_time.tv_sec * 1000000 + pgm_start_time.tv_usec);
	printf("\nmain() to Fini() time = %ld usec\n", elapsed_time);
	return;
}

int main(int argc, char *argv[])
{
	char filename[32];
	int myrank = 0, n_processes = 0;	

	gettimeofday(&pgm_start_time, NULL);

	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);	
	MPI_Comm_size(MPI_COMM_WORLD, &n_processes);
	printf("\n%d of %d", myrank, n_processes);
	sprintf(filename, "trace%d.txt", myrank);
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Finalize();

	PIN_Init(argc, argv);

	if((logfptr = fopen(filename, "w")) ==  NULL)
	{
		printf("trace file open error....exiting\n");
		exit(1);
	}	

	for(int i=0; i<MAX_THREADS; i++)
	{prev_ip[i] = 0;}

	// Activate instruction counter
	icount.Activate();
	icount_lastmem = 1;
	memrefs = 1;

	// Activate alarm, must be done before PIN_StartProgram
	control.CheckKnobs(Handler, 0);

	doinst = false;

	//PIN_AddThreadBeginFunction(ThreadBegin, 0);
	//PIN_AddThreadEndFunction(ThreadEnd, 0);
	//PIN_AddContextChangeFunction(ContextChangeCb, 0);
	PIN_InitSymbols(); 

	INS_AddInstrumentFunction(Instruction, 0);
	// Register Routine to be called to instrument rtn
	RTN_AddInstrumentFunction(Routine, 0);
	IMG_AddInstrumentFunction(ImageLoad, 0);
	PIN_AddFiniFunction(Fini, 0);

	// Never returns
	PIN_StartProgram();

	return 0;
}
