/*
 *  This file contains the implementation of EPRSD template library 
 *  Created on Dec 01 2009
 *  Author: Sandeep Budanur Ramanna
 *  E-mail: sbudanu@ncsu.edu
 *  North Carolina State University
 */

#include<stdio.h>
#include<math.h>
#include<iostream>
#include<stdlib.h>
#include "types.h"

#define SIGTREE_ENABLED
//#define IMPRECISION_ENABLED

#ifdef SIGTREE_ENABLED
#include "sigt.cpp"
#endif
//#include "myqueue.cpp"
using namespace std;

FILE* errfptr = NULL;

#define _MIN(a,b) ((a)<=(b))?(a):(b)
#define _MAX(a,b) ((a)>=(b))?(a):(b)
//#define _abs(a) ((a)>=0)?(a):-(a)
#define STDOUT stdout
#define DEFAULT_WINDOW_SIZE 1000

#define _SUCCESS 0
#define _ERROR -1

#define ATTR_NONE 1
#define ATTR_LEADER 2 //loop starting point
#define ATTR_MEMBER 4
#define ATTR_TAIL 8   //loop ending point

/* to differentiate inter-thread and inter-node merging */
#define INTERTHREAD_MERGE 0
#define INTERNODE_MERGE 1

#define MAX_SIGN_LEN 16

/* 100% match needed for merging */
#define PERCENT_MATCH (100.0)
#define LOOP_HEAD_SIGN 0xC0FFEEFEEDC0FFEE

typedef enum
{
	CODE_RSD=0,
	CODE_EPRSD
}CODE;

typedef enum
{
	VIRTUAL = 0,
	REAL		
}SIGN_TYPE;

const char* code_str[] = {"RSD", "EPRSD"};

/* COMMON class - contains mandatory info of EPRSDs*/
class COMMON
{
	public:
#ifdef SIGTREE_ENABLED
		void* sigtree_leaf;
#endif

	public:
		_UINT64* signature;
		//_UINT64 signature[MAX_SIGN_LEN];
		_UINT64 sign_len;
		_UINT64 max_sign_len;	
		_UINT64 ref_id;	
		_UINT64 sign_type; /* VIRTUAL if signature tree is used. REAL if actual signature list is sent for every EPRSD */
		_UINT64 type;
		_UINT64 thread_id;
		_UINT64 node_id;
		//INT64 value;
		char name[10];	

		/* constrcutor */
		COMMON(_UINT64 max_sign_len)
		{
			//cout<<endl<<"\nCOMMON ctor called";
			try
			{
				this->signature = new _UINT64[max_sign_len];
			}		
			catch(bad_alloc &e)
			{
				printf("\nnew in malloc failed, exit");
				exit(-10);
			}
			this->max_sign_len = max_sign_len;
			this->sign_len = 0;
			this->ref_id = 0;
#ifdef SIGTREE_ENABLED
			this->sign_type = VIRTUAL; /* reference to signature tree, not the actual signature list */
#else
			this->sign_type = REAL; /* actual signature list used */
#endif
			this->type = 0;
			this->thread_id = 0;
			this->node_id = 0;
			this->name[0] = '\0';
#ifdef SIGTREE_ENABLED
			this->sigtree_leaf = NULL;
#endif
		}

		COMMON(const COMMON& src)
		{
			_UINT64 i = 0;
			//cout<<endl<<"\nCOMMON copy ctor called";
			try
			{
				this->signature = new _UINT64[src.max_sign_len];
			}
			catch(bad_alloc &e)
			{
				printf("\nnew in malloc failed, exit");
				exit(-10);
			}
			this->max_sign_len = src.max_sign_len;
			this->sign_len = _MIN(src.max_sign_len, src.sign_len);
			this->sign_type = src.sign_type;

			for(i=0; i<this->sign_len; i++)
				signature[i] = src.signature[i];

			this->ref_id = src.ref_id;
			this->type = src.type;
			this->thread_id = src.thread_id;
			this->node_id = src.node_id;
			strcpy(this->name, src.name);
#ifdef SIGTREE_ENABLED
			this->sigtree_leaf = NULL;
#endif
		}

		~COMMON()
		{
			//printf("\nCOMMON dtor() - this = %p signature = %p", (void*)this, (void*)signature);
			if(signature)
				delete signature;
		}	

		void setData(_UINT64* signature, _UINT64 sign_len, _UINT64 ref_id, _UINT64 type, _UINT64 thread_id, _UINT64 node_id, const char* name)
		{
			int name_len = 0, j = 0;
			_UINT64 i = 0;

			//if(!this->signature){cerr<<"\nsign ptr NULL";}
			this->sign_len = _MIN(this->max_sign_len, sign_len);

			for(i=0; i<this->sign_len; i++)
				this->signature[i] = signature[i];

			this->ref_id = ref_id;
			this->type = type;
			this->thread_id = thread_id;
			this->node_id = node_id;
			name_len = _MIN(strlen(name),sizeof(this->name)-1);	
			for(j=0; j<name_len; j++)
				this->name[j] = name[j];
			this->name[j] = '\0';
			return;
		}

		void print(FILE* outfptr)
		{		
			_UINT64 i = 0;
#if 0
			fprintf(outfptr, "\n%ld ", sign_len);

			for(i=0; i<sign_len; i++)
				fprintf(outfptr, "%p ", (void*)signature[i]);

			fprintf(outfptr, "\n%p",(void*)ref_id);		
			fprintf(outfptr, "\n%p", (void*)type);
#else		
			fprintf(outfptr, "\nSign_len = %ld", sign_len);
			fprintf(outfptr, "\nSign:");

			for(i=0; i<sign_len; i++)
				fprintf(outfptr, "%p ", (void*)signature[i]);

			fprintf(outfptr, "\nref_id = %p",(void*)ref_id);		
			fprintf(outfptr, "\ntype = %p", (void*)type);
#endif
#if 0
			fprintf(outfptr, "\nthread_id = %ld", thread_id);
			fprintf(outfptr, "\nnode_id = %ld", node_id);
			fprintf(outfptr, "\nname = ");
			if(name)
				fprintf(outfptr, "%s", name);		
#endif
			return;
		}	
};

struct details
{
	_INT64 type; //e.g: TYPE_ADDR for address, TYPE_THREAD_ID, TYPE_NODE_ID
	/* start value, length and stride of intra-task values e.g: memory address, destination address, etc.*/	
	_INT64 start_value;
	_INT64 length;
	_INT64 stride;
	/* start value, length and stride of inter-task values (thread id) */
	_INT64 tid_start_value;
	_INT64 tid_length;
	_INT64 tid_stride;
	/* start value, length and stride of inter-node values (node id) */
	_INT64 node_id_start_value;
	_INT64 node_id_length;
	_INT64 node_id_stride;

	_INT64 loopsize;//no. of LD/ST in a loop iteration					
	_INT64 loopcount;//no. of loop iterations
	_INT64 eprsd_merge_count;
	_UINT64 tid_bit_pattern;
	_UINT64 nodeid_bit_pattern;
	_INT64 tid_addr_stride; /* address stride based on thread id (function of thread id) */
};

template <class T>
class EPRSD_DATA;

/* EPRSD class */
template <class T>
class EPRSD
{
	private:
		CODE code; //if beginning of a loop, it is CODE_EPRSD otherwise CODE_RSD
		_INT64 attr;
		struct details det[3];// if more than one stride exists

	public:
		EPRSD<T>* prev;
		EPRSD<T>* next;
		EPRSD<T>* endofloop;
		_INT64 gap;

	public:
		COMMON* cmninfo;//filled by user
		T* userinfo;//defined by user
		template <class U> friend class EPRSD_DATA;

		EPRSD(COMMON *cmninfo, T* userinfo)
		{
			this->cmninfo = new COMMON(*cmninfo);
			this->userinfo = new T(*userinfo);
			this->det[0].length = 1;
			this->det[0].stride = 0;
			//this->det[0].start_value = 0;
			this->det[0].loopsize = 0;
			this->det[0].loopcount = 0;
			this->det[0].eprsd_merge_count = 0;
			this->prev = NULL;
			this->next = NULL;
			this->endofloop = NULL;
			this->code = CODE_RSD;
			this->attr = ATTR_NONE;
			this->gap = 0;
			//this->cmninfo = cmninfo;
			//this->userinfo = userinfo;
			this->det[0].start_value = userinfo->getValue();
			this->det[0].tid_start_value = (this->cmninfo)->thread_id;
			this->det[0].tid_length = 1;
			this->det[0].tid_stride = 0;
			this->det[0].node_id_start_value = (this->cmninfo)->node_id;
			this->det[0].node_id_length = 1;
			this->det[0].node_id_stride = 0;
			this->det[0].tid_bit_pattern = 1<<(this->cmninfo)->thread_id;
			this->det[0].nodeid_bit_pattern = 1<<(this->cmninfo)->node_id;
			this->det[0].tid_addr_stride = 0;
		}

		EPRSD(EPRSD_DATA<T>* data)
		{
			int i=0;
			this->cmninfo = new COMMON(*(data->cmninfo));
			this->userinfo = new T(*(data->userinfo));

			for(i=0;i<data->det_len;i++)
			{
				this->det[i] = data->det[i];				
			}		
			this->prev = NULL;
			this->next = NULL;
			this->endofloop = NULL;
			this->code = data->code;
			this->attr = data->attr;
			this->gap = 0;
		}

		~EPRSD()
		{
			//printf("\n~EPRSD() = %p, cmn = %p, usr = %p", this, cmninfo, userinfo);
			if(cmninfo)	
				delete cmninfo;
			if(userinfo)
				delete userinfo;
		}

		EPRSD<T>* createLoopHead();
		EPRSD<T>* createLoopTail();

		void print(FILE* outfptr, _INT64* endloop_ref_id, _INT32* level_counter, _INT32* num_level)
		{
			//int i=0;
			//int tmp = 0;

			//printf("\n ptr = %p", this);
			//if(code == CODE_EPRSD)

#if 0	
			if(attr & ATTR_LEADER)
			{
				fprintf(outfptr, "\n\n<EPRSD BEGINS, LOOPSIZE: %ld LOOPCOUNT: %ld>", det[0].loopsize, det[0].length);
			}
#endif
			//fprintf(outfptr, "\n\ncode: %d", code);
#if 0
			fprintf(outfptr, "\n\n%ld %ld %ld", attr,  det[0].loopsize, det[0].length);
			fprintf(outfptr, "\n%ld", gap);
			fprintf(outfptr, "\n%p", (void*)det[0].start_value);
			fprintf(outfptr, "\n%ld", det[0].length);
			fprintf(outfptr, "\n%p", (void*)det[0].stride);
			fprintf(outfptr, "\n%p", (void*)det[0].tid_start_value);
			fprintf(outfptr, "\n%ld", det[0].tid_length);
			fprintf(outfptr, "\n%p", (void*)det[0].tid_stride);
			fprintf(outfptr, "\n%p", (void*)det[0].tid_bit_pattern);
			fprintf(outfptr, "\n%p", (void*)det[0].tid_addr_stride);
			fprintf(outfptr, "\n%p", (void*)det[0].node_id_start_value);
			fprintf(outfptr, "\n%ld", det[0].node_id_length);
			fprintf(outfptr, "\n%p", (void*)det[0].node_id_stride);
			fprintf(outfptr, "\n%p", (void*)det[0].nodeid_bit_pattern);
#else
			fprintf(outfptr, "\n\nattr: %ld loopsize: %ld loopcount: %ld, sign: %p", attr,  det[0].loopsize, det[0].length, cmninfo->ref_id);
			//fprintf(outfptr, "\n\nattr: %ld loopsize: %ld loopcount: %ld", attr,  det[0].loopsize, det[0].length);
			fprintf(outfptr, "\ngap: %ld", gap);
			fprintf(outfptr, "\nstart_value: %p", (void*)det[0].start_value);
			fprintf(outfptr, "\nlength: %ld", det[0].length);
			fprintf(outfptr, "\nstride: %p", (void*)det[0].stride);
			fprintf(outfptr, "\ntid_start_value: %p", (void*)det[0].tid_start_value);
			fprintf(outfptr, "\ntid_length: %ld", det[0].tid_length);
			fprintf(outfptr, "\ntid_stride: %p", (void*)det[0].tid_stride);
			fprintf(outfptr, "\ntid_bit_pattern: %p", (void*)det[0].tid_bit_pattern);
			fprintf(outfptr, "\ntid_addr_stride: %p", (void*)det[0].tid_addr_stride);
			fprintf(outfptr, "\nnode_id_start_value: %p", (void*)det[0].node_id_start_value);
			fprintf(outfptr, "\nnode_id_length: %ld", det[0].node_id_length);
			fprintf(outfptr, "\nnode_id_stride: %p", (void*)det[0].node_id_stride);
			fprintf(outfptr, "\nnode_id_bit_pattern: %p", (void*)det[0].nodeid_bit_pattern);
#endif
			//printf("\neprsd merge: %p", det[0].eprsd_merge_count+1);//n-1 merges produce n eprsd count

			if(cmninfo)			
				cmninfo->print(outfptr);

			/*
			   if((*num_level) > 0)	
			   {
			   printf("*%d*", *num_level);
			   if(endloop_ref_id[(*num_level)-1] == cmninfo->ref_id)
			   {
			   printf("\n<EPRSD ENDS: %p>", endloop_ref_id[(*num_level)-1]);
			   (*num_level)--; 
			   }	
			   }
			   */		

#if 0	 
			for(i=0;i<(*num_level);i++)
			{
				//printf("*%d*", *num_level);
				level_counter[i]--;
				if(level_counter[i] == 0)
				{					
					fprintf(outfptr, "\n</EPRSD ENDS: %p  LOOPCOUNT: %ld>",(void*)endloop_ref_id[i], det[0].length);
					//(*num_level)--;					
				}					
			}
			tmp = (*num_level);	
			for(i=0;i<tmp;i++)
			{
				if(level_counter[i] == 0)
					(*num_level)--;
			}
#endif			  
		}

		inline void setLength(_INT64 len)
		{ this->det[0].length = len; }
		inline void setStride(_INT64 stride)
		{ this->det[0].stride = stride; }
		inline void setStartValue(_INT64 start_value)
		{ this->det[0].start_value = start_value; }		

		inline _INT64 getLength()
		{ return this->det[0].length; }
		inline _INT64 getStride()
		{ return this->det[0].stride; }
		inline _INT64 getStartValue()
		{ return this->det[0].start_value; }

		inline void setCode(CODE code)
		{ this->code = code; }		
		inline CODE getCode()
		{ return this->code; }		

		inline void setAttr(_INT64 attr)
		{ this->attr |= attr; }		
		inline _INT64 getAttr()
		{ return this->attr; }

		inline _INT64 getLoopSize()
		{ return this->det[0].loopsize;	}
		inline void setLoopSize(_INT64 size)
		{ this->det[0].loopsize = size; }

		inline _INT64 getLoopCount()
		{ return this->det[0].loopcount;}
		inline void setLoopCount(_INT64 count)
		{ this->det[0].loopcount = count; }

		inline void setEprsdMergeCount(_INT64 count)
		{ this->det[0].eprsd_merge_count = count; }
		inline _INT64 getEprsdMergeCount()
		{ return this->det[0].eprsd_merge_count; }

		inline _INT64 getThreadId()
		{ return this->det[0].tid_start_value; }
		inline _INT64 getThreadIdLength()		
		{ return this->det[0].tid_length; }
		inline _INT64 getThreadIdStride()
		{ return this->det[0].tid_stride; }
		inline _UINT64 getThreadIdBits()
		{ return this->det[0].tid_bit_pattern; }
		inline _INT64 getThreadIdAddrStride()
		{ return this->det[0].tid_addr_stride; }

		inline void setThreadId(_INT64 tid)
		{ 
			this->det[0].tid_start_value = tid; 
			this->det[0].tid_bit_pattern |= 1<<tid;
		}
		inline void setThreadIdLength(_INT64 len)		
		{ this->det[0].tid_length = len; }
		inline void setThreadIdStride(_INT64 stride)
		{ this->det[0].tid_stride = stride; }
		inline void setThreadIdBits(_UINT64 bits)
		{  
			this->det[0].tid_bit_pattern |= bits;
		}
		inline void setThreadIdAddrStride(_INT64 stride)
		{ this->det[0].tid_addr_stride = stride; }

		inline _INT64 getNodeId()
		{ return this->det[0].node_id_start_value; }
		inline _INT64 getNodeIdLength()		
		{ return this->det[0].node_id_length; }
		inline _INT64 getNodeIdStride()
		{ return this->det[0].node_id_stride; }
		inline _UINT64 getNodeIdBits()
		{ return this->det[0].nodeid_bit_pattern; }

		inline void setNodeId(_INT64 id)
		{ 
			this->det[0].node_id_start_value = id; 
			this->det[0].nodeid_bit_pattern |= 1<<id;
		}
		inline void setNodeIdLength(_INT64 len)		
		{ this->det[0].node_id_length = len; }
		inline void setNodeIdStride(_INT64 stride)
		{ this->det[0].node_id_stride = stride; }
		inline void setNodeIdBits(_UINT64 bits)
		{  
			this->det[0].nodeid_bit_pattern |= bits;
		}
};

template <class T>
class EPRSD_DATA
{
	public:
		CODE code; //if beginning of a loop, it is CODE_EPRSD otherwise CODE_RSD
		_INT64 attr;
		_INT16 det_len; //no. of details structure used
		struct details det[3];// if more than one stride exists	
		COMMON* cmninfo;//filled by user
		T* userinfo;//defined by user

	public:
		EPRSD_DATA()
		{
			code = CODE_RSD;
			attr = ATTR_NONE;
			det_len = 1;
			cmninfo = NULL;
			userinfo = NULL;	
		}

		//creates EPRSD_DATA from EPRSD class
		void copy(EPRSD<T> *src)
		{
			code = src->code;
			attr = src->attr;	
			det_len = 1; //TODO: modify this
			det[0] = src->det[0];
			det[1] = src->det[1];
			det[2] = src->det[2];
			//only pointer is copied
			cmninfo = src->cmninfo;
			userinfo = src->userinfo;			
		}

		//creates EPRSD_DATA from details, cmn and userinfo
		void copy(CODE code, _INT64 attr, struct details* det, int det_len, COMMON* cmninfo, T* userinfo)
		{
			int i = 0;
			this->code = code;
			this->attr = attr;
			for(i=0;i<det_len;i++)
			{
				this->det[i] = det[i];
			}
			//data is copied, so allocate memory
			this->cmninfo = new COMMON(*cmninfo);
			this->userinfo = new T(*userinfo);
		}		
};

/* EPRSD Compressor class*/
template <class T, template <class M> class MATCH, template <class G> class MERGE>
class EPRSD_COMPRESSOR
{
	private:
		EPRSD<T> *listhead;
		EPRSD<T> *listtail;
		/* used for inter-thread an inter-node merging */
		EPRSD<T> *rlisthead;
		EPRSD<T> *rlisttail;

		EPRSD<T> *l_match_begin;//starting node of left matching portion 
		EPRSD<T> *l_match_end;//ending node of left matching portion	
		EPRSD<T> *r_match_begin;//starting node of right matching portion 
		EPRSD<T> *r_match_end;//ending node of right matching portion
		//EPRSD<T> *latest_eprsd;//ending node of right matching portion
		/* for inter task compression */
		EPRSD<T> *it_l_match_begin;
		EPRSD<T> *it_l_match_end;
		EPRSD<T> *it_r_match_begin;
		EPRSD<T> *it_r_match_end;
		EPRSD<T> *it_prev_match;
		EPRSD<T> *in_l_match_begin;
		EPRSD<T> *in_l_match_end;
		EPRSD<T> *in_r_match_begin;
		EPRSD<T> *in_r_match_end;
		EPRSD<T> *in_prev_match;
		EPRSD<T> *prev_merge_point;		
#ifdef SIGTREE_ENABLED
		SIGTREE *sigtree;
		SIGNODE *loophead_leaf;
#endif
		_UINT64 max_window_size;
		_UINT64 rank;	

	public:
		_INT32 level_counter[10];
		_INT64 endloop_ref_id[10];
		_INT32 num_level;

	private:
		template <class _T, template <class _M> class _MATCH, template <class _G> class _MERGE>
			friend class ITERATOR;

		/* returns the pointer to matched EPRSD otherwise NULL */
		bool isSignMatch(COMMON* src, COMMON* dest)
		{
			_UINT64 i=0;
			if(src == NULL || dest == NULL)
			{
				printf("\nError: NULL COMMON ptrs");
				exit(-4);				
			}
			if(src->ref_id != dest->ref_id)
				return false;
			else //compare signature
			{
				if(src->sign_len != dest->sign_len)
					return false;
				else
				{
					//printf("\nsrc = %p dest = %p Sign_len = %ld", src, dest, src->sign_len);
					//printf("\nsrc->sign = %p dest->sign = %p", src->signature, dest->signature);
					for(i=0;i<src->sign_len;i++)
					{	
						if(src->signature[i] != dest->signature[i])
							return false;
					}
				}
			}
			return true;
		}

		/* match the tail node with the other nodes,  loose match needed for add() - compare 
		   only sign and userinfo - PRSD/EPRSD are NOT differentiated */
		EPRSD<T>* findTailMatch()
		{
			EPRSD<T>* neweprsd = listtail;
			EPRSD<T>* cur = NULL;
			_UINT64 cmp_window_size = max_window_size;

			//printf("\nfindTailMatch - %p", listtail);
			//list empty
			if(!listhead || !listtail || !neweprsd)
				return NULL;

			cur = listtail->prev;
			while(cur && cmp_window_size)
			{
				//printf("\nFLM ->%p", cur);
				//if(cur->getCode() == CODE_RSD && neweprsd->getCode() == CODE_RSD)//compare only RSDs, RSD and EPRSD don't match
				{
					//if( match_fun(cur->cmninfo, cur->userinfo, neweprsd->cmninfo, neweprsd->userinfo) )
#ifdef SIGTREE_ENABLED
					if( 	cur->cmninfo->sigtree_leaf == neweprsd->cmninfo->sigtree_leaf\
							&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) )
#else
						if( 	isSignMatch(cur->cmninfo, neweprsd->cmninfo)\
								&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) )
#endif	
						{					
							return cur;
						}
				}
				cur = cur->prev;
				cmp_window_size--;	
			}
			return NULL;
		}
#if 0
		/* loose match needed for add() - compare only sign and userinfo - PRSD/EPRSD are NOT differentiated*/
		EPRSD<T>* findLastMatch(EPRSD<T>* neweprsd)
		{
			EPRSD<T>* cur = NULL;
			//EPRSD<T>* next = NULL;
			_UINT64 cmp_window_size = max_window_size;
			//list empty
			if(!listhead || !listtail || !neweprsd)
				return NULL;

			cur = listtail;
			while(cur && cmp_window_size)
			{
				//printf("\nFLM ->%p", cur);
				//if(cur->getCode() == CODE_RSD && neweprsd->getCode() == CODE_RSD)//compare only RSDs, RSD and EPRSD don't match
				{
					//if( match_fun(cur->cmninfo, cur->userinfo, neweprsd->cmninfo, neweprsd->userinfo) )
#ifdef SIGTREE_ENABLED
					if( 	cur->cmninfo->sigtree_leaf == neweprsd->cmninfo->sigtree_leaf\
							&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) )
#else
						if( 	isSignMatch(cur->cmninfo, neweprsd->cmninfo)\
								&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) )
#endif	
						{					
							return cur;
						}
				}
				cur = cur->prev;
				cmp_window_size--;	
			}
			return NULL;
		}
#endif

		/* tight match needed  - compare sign, userinfo, type, loop size. PRSD/EPRSD are differentiated unlike the findLastMatch()*/
		EPRSD<T>* findExactMatch(EPRSD<T>* neweprsd, _INT64 *gap)
		{
			EPRSD<T>* cur = NULL;
			*gap = 0;
			//EPRSD<T>* next = NULL;
			//fprintf(STDOUT, "\nfindExactMatch called - %p", neweprsd);
			//list empty
			if(!listhead || !listtail || !neweprsd)
				return NULL;
			cur = neweprsd->next;
			while(cur)
			{				
				//if( match_fun(cur->cmninfo, cur->userinfo, neweprsd->cmninfo, neweprsd->userinfo) &&
#ifdef SIGTREE_ENABLED
				if( 	cur->cmninfo->sigtree_leaf == neweprsd->cmninfo->sigtree_leaf \
						&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) &&
						cur->getAttr() == neweprsd->getAttr() &&
						(cur->getLoopSize() == neweprsd->getLoopSize())
				  )
#else
					if( 	isSignMatch(cur->cmninfo, neweprsd->cmninfo)\
							&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) &&
							cur->getAttr() == neweprsd->getAttr() &&
							(cur->getLoopSize() == neweprsd->getLoopSize())
					  )
#endif 
						//  (((cur->endofloop)->cmninfo)->ref_id == ((neweprsd->endofloop)->cmninfo)->ref_id)					
					{
						//fprintf(STDOUT, "\nfindExactMatch gap = %d %d", gap, neweprsd->getEprsdMergeCount());
						//neweprsd->gap = (neweprsd->gap + gap)/(neweprsd->getEprsdMergeCount() + 1);
						return cur;
					}
				(*gap)++;
				cur = cur->next;
			}
			//printf("$3");
			return NULL;
		}

		EPRSD<T>* internode_findExactMatch(EPRSD<T>* neweprsd)
		{			
			EPRSD<T>* cur = NULL;
			//EPRSD<T>* next = NULL;
			//fprintf(STDOUT, "\nfindExactMatch called - %p", neweprsd->cmninfo->ref_id);
			//list empty
			if(!listhead || !listtail || !neweprsd)
				return NULL;

			/* need not search from beginning every time, resume from previous match onwards */
			/*
			   if(it_l_match_begin && it_l_match_end)//if window already formed
			   cur = it_l_match_end->next;
			   else
			   cur = listhead;
			   */
#if 1 /* Inter-task/Inter-node compression June 11, 2010 bug */
			in_prev_match = prev_merge_point;
			if(in_prev_match == NULL)
				cur = listhead;
			else 
				cur = in_prev_match->next;
#else
			cur = listhead;
#endif
			while(cur)
			{
#ifdef SIGTREE_ENABLED
				if( cur->cmninfo->sigtree_leaf == neweprsd->cmninfo->sigtree_leaf && 
						MATCH<T>::internode_compare(cur->userinfo, neweprsd->userinfo) &&
						cur->getAttr() == neweprsd->getAttr() &&
						(cur->getLoopSize() == neweprsd->getLoopSize()) &&
						(cur->getLoopCount() == neweprsd->getLoopCount()) &&
						(cur->getStride() == neweprsd->getStride())&&
						(cur->getThreadIdLength() == neweprsd->getThreadIdLength())&&
						(cur->getThreadIdStride() == neweprsd->getThreadIdStride())&&
						(cur->getThreadId() == neweprsd->getThreadId()) &&
						(cur->getNodeIdLength() == neweprsd->getNodeIdLength())&&
						(cur->getNodeIdStride() == neweprsd->getNodeIdStride())
				  ) 
#else
					if( isSignMatch(cur->cmninfo, neweprsd->cmninfo)&& 
							MATCH<T>::internode_compare(cur->userinfo, neweprsd->userinfo) &&
							cur->getAttr() == neweprsd->getAttr() &&
							(cur->getLoopSize() == neweprsd->getLoopSize()) &&
							(cur->getLoopCount() == neweprsd->getLoopCount()) &&
							(cur->getStride() == neweprsd->getStride()) &&
							(cur->getThreadIdLength() == neweprsd->getThreadIdLength())&&
							(cur->getThreadIdStride() == neweprsd->getThreadIdStride())&&
							(cur->getThreadId() == neweprsd->getThreadId()) &&
							(cur->getNodeIdLength() == neweprsd->getNodeIdLength())&&
							(cur->getNodeIdStride() == neweprsd->getNodeIdStride())
					  ) 
#endif
					{
						//fprintf(STDOUT, "\ninternode match found #1");
						//fprintf(STDOUT, "$$$$%ld %ld, %ld %ld, %ld %ld$$$$\n", cur->getThreadIdLength(), neweprsd->getThreadIdLength(),
						//	cur->getThreadIdStride(), neweprsd->getThreadIdStride(), cur->getThreadId(), neweprsd->getThreadId());

						if(cur->getThreadId() != neweprsd->getThreadId())
							fprintf(STDOUT, "$$$TID NOT EQUAL$$$");
						if(cur->getThreadIdLength() != neweprsd->getThreadIdLength())
							fprintf(STDOUT, "$$$TIDLEN NOT EQUAL$$$");
						if(cur->getThreadIdStride() != neweprsd->getThreadIdStride())
							fprintf(STDOUT, "$$$TIDSTRIDE NOT EQUAL$$$");

						break;
					}
				cur = cur->next;
			}
			if(cur)
				in_prev_match = cur; /* store the latest match so that we can search from this point onwards in the subsequent calls */	
			return cur;
		}

		EPRSD<T>* interthread_findExactMatch(EPRSD<T>* neweprsd)
		{
#if 1
			EPRSD<T>* cur = NULL;
			//EPRSD<T>* next = NULL;
			//fprintf(STDOUT, "\nfindExactMatch called - %p", neweprsd->cmninfo->ref_id);
			//list empty
			if(!listhead || !listtail || !neweprsd)
				return NULL;

			/* need not search from beginning every time, resume from previous match onwards */
			/*
			   if(it_l_match_begin && it_l_match_end)//if window already formed
			   cur = it_l_match_end->next;
			   else
			   cur = listhead;
			   */

#if 1 //IT mergeing bug - June 11, 2010
			it_prev_match = prev_merge_point;
			if(it_prev_match == NULL)
				cur = listhead;
			else 
				cur = it_prev_match->next;
#else
			cur = listhead;
#endif
			while(cur)
			{
#ifdef SIGTREE_ENABLED
				if( cur->cmninfo->sigtree_leaf == neweprsd->cmninfo->sigtree_leaf && 
						MATCH<T>::interthread_compare(cur->userinfo, neweprsd->userinfo) &&
						cur->getAttr()  == neweprsd->getAttr() &&
						(cur->getLoopSize() == neweprsd->getLoopSize()) &&
						(cur->getLoopCount() == neweprsd->getLoopCount()) &&
						(cur->getStride() == neweprsd->getStride())
						//&&
						//(cur->getThreadIdLength() == neweprsd->getThreadIdLength())&&
						//(cur->getThreadIdStride() == neweprsd->getThreadIdStride())
				  ) 
#else
					if( isSignMatch(cur->cmninfo, neweprsd->cmninfo)&& 
							MATCH<T>::interthread_compare(cur->userinfo, neweprsd->userinfo) &&
							cur->getAttr() == neweprsd->getAttr() &&
							(cur->getLoopSize() == neweprsd->getLoopSize()) &&
							(cur->getLoopCount() == neweprsd->getLoopCount()) &&
							(cur->getStride() == neweprsd->getStride())
							//&&
							//(cur->getThreadIdLength() == neweprsd->getThreadIdLength())&&
							//(cur->getThreadIdStride() == neweprsd->getThreadIdStride())
					  ) 
#endif
					{
						//fprintf(STDOUT, "\nintertask match found #1");
						break;
					}
					else
					{
#if 0
						printf("\n<<<<ref id %p %p >>>>", (void*)cur->cmninfo->ref_id, (void*)neweprsd->cmninfo->ref_id);
#ifdef SIGTREE_ENABLED
						if( cur->cmninfo->sigtree_leaf != neweprsd->cmninfo->sigtree_leaf)
						{
							printf("\nLEAF NOT EQUAL");
						}	
#endif
						if( !(MATCH<T>::intertask_compare(cur->userinfo, neweprsd->userinfo)) )
						{
							printf("\nUSERINFO NOT EQUAL");
						}	
						if(cur->getCode() != neweprsd->getCode())
						{
							printf("\nCODE NOT EQUAL");
						} 	
						if(cur->getLoopSize() != neweprsd->getLoopSize())
						{
							printf("\nLOOP SIZE NOT EQUAL");
						}	
#endif
					}	
				cur = cur->next;
			}
#endif
			if(cur)
				it_prev_match = cur; /* store the latest match so that we can search from this point onwards in the subsequent calls */	
			return cur;
		}

		void removeNoder(EPRSD<T>* neweprsd)
		{
			if(!neweprsd)
			{
				fprintf(stderr, "\nError: removeNoder() - null ptr");
				return;
			}
			if(!rlisthead)
			{
				fprintf(stderr, "\nError: removeNoder() - empty list");
				return;
			}

			if(neweprsd == rlisthead && neweprsd == rlisttail) //onle one node
			{
				rlisthead  = rlisttail = NULL;
			}	
			else if(neweprsd == rlisthead) //head node removal
			{
				rlisthead = rlisthead->next;
				rlisthead->prev = NULL;
				neweprsd->next = NULL;
			}
			else if(neweprsd == rlisttail) //tail node removal
			{
				rlisttail = rlisttail->prev;
				rlisttail->next = NULL;
				neweprsd->prev = NULL;
			}
			else //middle node removal
			{
				EPRSD<T>* prv = neweprsd->prev;
				EPRSD<T>* nxt = neweprsd->next;
				prv->next = nxt;
				nxt->prev = prv;
				neweprsd->prev = neweprsd->next = NULL;	
			}
			delete neweprsd;
			return;
		}

		void addTailr(EPRSD<T>* neweprsd)
		{
			static int count = 0;
			if(!neweprsd)
				return;
			if(!rlisthead)
			{
				rlisthead = neweprsd;
				rlisttail = neweprsd;
				neweprsd->prev = NULL;
				neweprsd->next = NULL;
			}
			else
			{
				rlisttail->next = neweprsd;
				neweprsd->prev = rlisttail;
				rlisttail = neweprsd;
				neweprsd->next = NULL;
			}
			count++;
			//printf("\nadded %d nodes", count);
			return;
		}

		void addHeadr(EPRSD<T>* neweprsd)
		{
			if(!neweprsd)
				return;
			if(!rlisthead)
			{
				rlisthead = neweprsd;
				rlisttail = neweprsd;
				neweprsd->prev = NULL;
				neweprsd->next = NULL;
			}
			else
			{
				rlisthead->prev = neweprsd;
				neweprsd->next = rlisthead;
				rlisthead = neweprsd;
				neweprsd->prev = NULL;
			}
			return;
		}


		void addTail(EPRSD<T>* neweprsd)
		{
			if(!neweprsd)
				return;
			//neweprsd->setCode(CODE_RSD);
			if(!listhead)
			{
				listhead = neweprsd;
				listtail = neweprsd;
				neweprsd->prev = NULL;
				neweprsd->next = NULL;
			}
			else
			{
				listtail->next = neweprsd;
				neweprsd->prev = listtail;
				listtail = neweprsd;
				neweprsd->next = NULL;
			}
			return;
		}

		void addHead(EPRSD<T>* neweprsd)
		{
			if(!neweprsd)
				return;
			//neweprsd->setCode(CODE_RSD);
			if(!listhead)
			{
				listhead = neweprsd;
				listtail = neweprsd;
				neweprsd->prev = NULL;
				neweprsd->next = NULL;
			}
			else
			{
				listhead->prev = neweprsd;
				neweprsd->next = listhead;
				listhead = neweprsd;
				neweprsd->prev = NULL;
			}
			return;
		}

		void interthread_merge(EPRSD<T>*, EPRSD<T>*);
		void interthread_mergeLeftRightPortions();		
		void interthread_addSequenceAfter(EPRSD<T>* prev_merge_point);
		void internode_merge(EPRSD<T>*, EPRSD<T>*);
		void internode_mergeLeftRightPortions();
		_INT64 mergeLeftRightPortions();
		void addAfter(EPRSD<T>* neweprsd, EPRSD<T>* prev_merge_point);
		void insertSequenceAfter(EPRSD<T>* prev_merge_point, EPRSD<T>* start, EPRSD<T>* end);			
		void printWindow(FILE* outfptr, EPRSD<T>* start, EPRSD<T>* end)
		{
			EPRSD<T>* cur;
			//fprintf(STDOUT, "\nprintWindow\n");
			if(!start || !end)
				return;
			cur = start;
			while(cur && cur != end->next)
			{
				//printf("\n*%p*", cur);
				cur->print(outfptr);
				cur = cur->next;
			}			
		}		

	public:
		//template <class T_, template <class M_> class MATCH_, template <class G_> class MERGE_>
		class ITERATOR
		{
			private:
				EPRSD_COMPRESSOR *cmprsr;
				EPRSD<T> *cur;
				EPRSD_DATA<T> *data;	

			public:
				ITERATOR(EPRSD_COMPRESSOR<T, MATCH, MERGE>* ecr)
				{
					cmprsr = ecr;
					cur = NULL;
					data = new EPRSD_DATA<T>();
				}

				~ITERATOR()
				{
					delete data;
				}	

				void begin()
				{
					cur = cmprsr->listhead;
				}

				EPRSD_DATA<T>* getNext()
				{
					EPRSD<T> *tmp = NULL;	
					tmp = cur;		
					if(cur)
						cur = cur->next;		
					//return tmp;
					if(tmp)
					{
						data->copy(tmp);//copy EPRSD data portion
						return data;
					}
					return NULL;
				}
		};

		EPRSD_COMPRESSOR(_UINT64 rank)
		{
			int i=0;
			this->listhead = NULL;
			this->listtail = NULL;				
			this->rlisthead = NULL;
			this->rlisttail = NULL;				
			this->l_match_begin = NULL;
			this->l_match_end = NULL;
			this->r_match_begin = NULL;
			this->r_match_end = NULL;
			this->it_l_match_begin = NULL;
			this->it_l_match_end = NULL;
			this->it_r_match_begin = NULL;
			this->it_r_match_end = NULL;
			this->it_prev_match = NULL;
			//this->latest_eprsd = NULL;
			this->num_level = 0;
			this->in_l_match_begin = NULL;
			this->in_l_match_end = NULL;
			this->in_r_match_begin = NULL;
			this->in_r_match_end = NULL;
			this->in_prev_match = NULL;
			for(i=0;i<10;i++)
			{
				this->level_counter[i] = 0;
				this->endloop_ref_id[i] = 0;
			}
#ifdef SIGTREE_ENABLED
			this->sigtree = new SIGTREE();
			sigtree->initAdding();
			this->loophead_leaf = (SIGNODE*)sigtree->add((void*)LOOP_HEAD_SIGN); /* default node for loop head */ 
			sigtree->finishAdding();
#endif
			this->max_window_size = DEFAULT_WINDOW_SIZE;

			if(STDOUT == NULL)
			{
				printf("\nSTDOUT file pointer is NULL");
				exit(3);
			}
			this->rank = rank;
		}

		~EPRSD_COMPRESSOR()
		{
			EPRSD<T>* cur = NULL;
			EPRSD<T>* nxt = NULL;

			if(!listhead)	
				return;

			cur = listhead;
			while(cur)
			{
				nxt = cur->next;
				delete cur;
				cur = nxt;
			}
			listhead = NULL;

			if(!rlisthead)	
				return;

			cur = rlisthead;
			while(cur)
			{
				nxt = cur->next;
				delete cur;
				cur = nxt;
			}
			rlisthead = NULL;

#ifdef SIGTREE_ENABLED
			delete sigtree;
#endif
		}

		void add(COMMON* cmn, T* newentry);
		void interthread_add(EPRSD_DATA<T> *data);
		void internode_add(EPRSD_DATA<T> *data);
		void startMerge(_INT64 merge_type);
		void interthread_startMerge();
		void internode_startMerge();
		void finalize();
		inline void setNodeId(_UINT64 nodeid)
		{
			EPRSD<T>* cur = NULL;

			if(!listhead)	
				return;

			fprintf(stdout, "\nsetNodeId %ld", nodeid);

			cur = listhead;

			while(cur)
			{	
				cur->setNodeId((_INT64)(1<<nodeid));
				cur = cur->next;
			}
			return;
		}

#ifdef SIGTREE_ENABLED
		SIGTREE_ITERATOR* createSigTreeIterator();
		void releaseSigTreeIterator(SIGTREE_ITERATOR* iter);
		void mergeToSigTree(SIGTREEITEM* iter);
		void numberSigTree(_UINT64 tid, _UINT64 nodeid);
		void assignUniqueKeys();
		bool mergeSigTree(EPRSD_COMPRESSOR<T, MATCH, MERGE> *src_ecr);
#endif	

		/* reset the required data structures before new set of EPRSDs arrive */
		void newEPRSDSetBegins()
		{
			this->it_prev_match = NULL;
			this->it_l_match_begin = NULL;
			this->it_l_match_end = NULL;
			this->it_r_match_begin = NULL;
			this->it_r_match_end = NULL;
			this->in_prev_match = NULL;
			this->in_l_match_begin = NULL;
			this->in_l_match_end = NULL;
			this->in_r_match_begin = NULL;
			this->in_r_match_end = NULL;
			this->prev_merge_point = NULL;
		}


		void print(FILE* outfptr)
		{
			EPRSD<T>* cur = NULL;
			if(!listhead)	
				return;
			cur = listhead;
			while(cur)
			{				
				cur->print(outfptr, endloop_ref_id, level_counter, &num_level);	
				cur = cur->next;	
			}
#if 0
			fprintf(outfptr, "\n/////////////////R LIST/////////////////\n");
			if(!rlisthead)	
			{
				fprintf(outfptr, "\n/////////////////EMPTY/////////////////\n");
				//return;
			}
			cur = rlisthead;
			while(cur)
			{				
				cur->print(outfptr, endloop_ref_id, level_counter, &num_level);	
				cur = cur->next;	
			}
			fprintf(outfptr, "\n////////////////////////////////////////\n");
#endif	

#ifdef SIGTREE_ENABLED
			/* printed on stdout */
			fprintf(STDOUT, "\nSIGTREE\n");
			sigtree->print(STDOUT);
#endif
		}

		inline void setCompareWindowSize(_UINT64 size)
		{
			max_window_size = size;
		}

		inline _UINT64 getCompareWindowSize()
		{
			return max_window_size;
		}
};

#ifdef SIGTREE_ENABLED
	template <class T, template <class M> class MATCH, template <class G> class MERGE>
SIGTREE_ITERATOR* EPRSD_COMPRESSOR<T, MATCH, MERGE>::createSigTreeIterator()
{
	SIGTREE_ITERATOR *iter = NULL;
	iter = sigtree->createIterator();
	return iter;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::releaseSigTreeIterator(SIGTREE_ITERATOR* iter)
{
	sigtree->releaseIterator(iter);

}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::mergeToSigTree(SIGTREEITEM* iter)
{
	sigtree->merge(iter);
	return;	
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::numberSigTree(_UINT64 tid, _UINT64 nodeid)
{
	_UINT64 num = 0; /* tree numbering begins with 0 */
	sigtree->numberTree(&num, tid, nodeid);
	printf("\n%ld nodes in %ld-%ld signature tree", num, tid, nodeid);
	return;
}

/* assigns EPRSDs the unique keys of the corresponding leaf nodes in the signature tree. This method *MUST* be called *AFTER* numberKeys() */
	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::assignUniqueKeys()
{
	EPRSD<T>* cur = NULL;
	SIGNODE* leaf = NULL;
	if(!listhead)	
		return;

	cur = listhead;

	while(cur)
	{
		cur->cmninfo->sign_len = 4;
		leaf = (SIGNODE*)cur->cmninfo->sigtree_leaf;
		if(!leaf)
		{
			fprintf(STDOUT, "\nError: assignUniqueKeys - NULL leaf ptr");
		}
		else
		{
			cur->cmninfo->signature[0] = (_UINT64)leaf->data.info;
			cur->cmninfo->signature[1] = leaf->data.key[0];
			cur->cmninfo->signature[2] = leaf->data.key[1];
			cur->cmninfo->signature[3] = leaf->data.key[2];
		}
		cur = cur->next;
	}

	return;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
bool EPRSD_COMPRESSOR<T, MATCH, MERGE>::mergeSigTree(EPRSD_COMPRESSOR<T, MATCH, MERGE> *src_ecr)
{
	sigtree->merge(src_ecr->sigtree);
	return true;
}

#endif

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::interthread_addSequenceAfter(EPRSD<T>* prev_merge_point)
{
	if(!prev_merge_point) /* insert in the beginning */
	{
		fprintf(STDOUT, "!!!!! IN - sequence inserted at the beginning, possible error");
		listhead = it_r_match_begin;
		listtail = it_r_match_end;
		it_r_match_begin->prev = NULL;
		it_r_match_end->next = NULL;
	}
	else if(prev_merge_point == listtail)
	{
		fprintf(STDOUT, "Error: IN - sequence cannot be inserted after tail...unlikely scenario");
		exit(-6);
	}
	else
	{
		it_r_match_begin->prev->next = it_r_match_end->next;
		if(it_r_match_end->next)
			it_r_match_end->next->prev = it_r_match_begin->prev;
		if(listtail == it_r_match_end)
		{
			listtail = it_r_match_begin->prev;
			listtail->next = NULL;
		}

		it_r_match_begin->prev = NULL;
		it_r_match_end->next = NULL;

		/* add after prev_merge_point */
		it_r_match_begin->prev = prev_merge_point;
		it_r_match_end->next = prev_merge_point->next;
		prev_merge_point->next->prev = it_r_match_end;
		prev_merge_point->next = it_r_match_begin;
	}
	return;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::insertSequenceAfter(EPRSD<T>* prev_merge_point, EPRSD<T>* start, EPRSD<T>* end)
{
	if(!start || !end)
	{
		fprintf(STDOUT, "\nIN(%d):insertSeq - NULL start/end ptrs", rank);
		return;
	}
	if(!prev_merge_point)//add in the very beginning
	{
		//fprintf(STDOUT, "\nIN(%d):insertSeq - in the very beginning", rank);
		if(listhead)
		{
			end->next = listhead;
			listhead->prev = end;
			listhead = start;
			listhead->prev = NULL;
		}
		else //empty list
		{
			listhead = start;
			listtail = end;
			listhead->prev = NULL;
			listtail->next = NULL;	
		}
	}
	else if(prev_merge_point == listtail)//in the tail end
	{
		//fprintf(STDOUT, "\nIN(%d):insertSeq - in the tail end", rank);
		listtail->next = start;
		start->prev = listtail;
		listtail = end;
		listtail->next = NULL;
	}
	else //in the middle
	{
		//fprintf(STDOUT, "\nIN(%d):insertSeq - in the middle", rank);
		start->prev = prev_merge_point;
		end->next = prev_merge_point->next;
		prev_merge_point->next = start;
		end->next->prev = end;
	}




#if 0
	if(!prev_merge_point) /* insert in the beginning */
	{
		fprintf(STDOUT, "prev merge point is NULL");

		return;
	}
	else if(prev_merge_point == listtail)
	{
		fprintf(STDOUT, "Error: IN - sequence cannot be inserted after tail...unlikely scenario");
		return;
	}
	else
	{
		in_r_match_begin->prev->next = in_r_match_end->next;
		if(in_r_match_end->next)
			in_r_match_end->next->prev = in_r_match_begin->prev;
		if(listtail == in_r_match_end)
		{
			listtail = in_r_match_begin->prev;
			listtail->next = NULL;
		}

		in_r_match_begin->prev = NULL;
		in_r_match_end->next = NULL;

		/* add after prev_merge_point */
		in_r_match_begin->prev = prev_merge_point;
		in_r_match_end->next = prev_merge_point->next;
		prev_merge_point->next->prev = in_r_match_end;
		prev_merge_point->next = in_r_match_begin;
	}
#endif
	return;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::addAfter(EPRSD<T>* neweprsd, EPRSD<T>* prev_merge_point)
{
	if(!prev_merge_point) /* add as head */
	{
		neweprsd->prev = NULL;
		neweprsd->next = listhead;
		if(listhead)
			listhead->prev = neweprsd;
		listhead = neweprsd;		
	}
	else if(listtail == prev_merge_point) /* add as tail */
	{
		addTail(neweprsd);
	}
	else
	{
		neweprsd->prev = prev_merge_point;
		neweprsd->next = prev_merge_point->next;
		prev_merge_point->next->prev = neweprsd;
		prev_merge_point->next = neweprsd;
	}
	return;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::internode_mergeLeftRightPortions()
{
	EPRSD<T> *cur = NULL;
	EPRSD<T> *end = NULL;
	EPRSD<T> *right_cur = NULL;
	EPRSD<T> *nxt = NULL;
	//_INT64 stride = 0;
	//_INT64 length = 0;
	_INT64 count = 0;
	//EPRSD is already present
	if(in_l_match_begin->getAttr() & ATTR_LEADER && in_l_match_end->getAttr() & ATTR_TAIL)//beginning of loop already marked, so just add the length
	{
#if 1
		//update the length of RSDs
		cur = in_l_match_begin;
		right_cur = in_r_match_begin;
		while(cur != in_l_match_end->next)
		{
			internode_merge(cur, right_cur);
			//MERGE<T>::internode_mergecb(cur->userinfo, right_cur->userinfo);
			cur = cur->next;
			right_cur = right_cur->next;
			//compute the EPRSD length, # of RSDs
			count++;
		}

		if(in_l_match_begin->getLoopSize() != count)
		{  
			//fprintf(STDOUT, "ERROR: internode EPRSD merging - count mismatch\n");	
		}	

		//adjust the head/tail before deleting right portion		
		//l_match_end->next = r_match_end->next;
		if(listtail == in_r_match_end)
		{
			listtail = in_r_match_begin->prev;
			listtail->next = NULL;
		}
		else /* needed for partial match merging */
		{
			in_r_match_begin->prev->next = in_r_match_end->next;
			in_r_match_end->next->prev = in_r_match_begin->prev;
			in_r_match_begin->prev = NULL;
			in_r_match_end->next = NULL;	
		}
#endif
		//TODO: fix this mem free crash
#if 1
		//free the right portion
		cur = in_r_match_begin;
		end = in_r_match_end->next;
		while(cur != end && cur)
		{
			nxt = cur->next;
			delete cur;
			cur = nxt;	
		}
		//fprintf(STDOUT, "\nMerge LR Update");		
#endif
	}
	return;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::interthread_mergeLeftRightPortions()
{
	EPRSD<T> *cur = NULL;
	EPRSD<T> *end = NULL;
	EPRSD<T> *right_cur = NULL;
	EPRSD<T> *nxt = NULL;	
	//_INT64 stride = 0;
	//_INT64 length = 0;
	_INT64 count = 0;	
	//EPRSD is already present
	if(it_l_match_begin->getAttr() & ATTR_LEADER && it_l_match_end->getAttr() & ATTR_TAIL)//beginning of loop already marked, so just add the length
	{
#if 1
		//printf("\nintertask_mergeLeftRightPortions - merging two EPRSDs");
		//update the length of RSDs
		cur = it_l_match_begin;
		right_cur = it_r_match_begin;
		while(cur != it_l_match_end->next)
		{
			interthread_merge(cur, right_cur);
			//MERGE<T>::intertask_mergecb(cur->userinfo, right_cur->userinfo);
			cur = cur->next;
			right_cur = right_cur->next;
			//compute the EPRSD length, # of RSDs
			count++;
		}

		if(it_l_match_begin->getLoopSize() != count)
		{  
			//fprintf(STDOUT, "ERROR: intertask EPRSD merging - count mismatch ");	
		}	

		//adjust the head/tail before deleting right portion		
		//l_match_end->next = r_match_end->next;
		if(listtail == it_r_match_end)
		{
			listtail = it_r_match_begin->prev;
			listtail->next = NULL;
		}
		else /* needed for partial match merging */
		{
			it_r_match_begin->prev->next = it_r_match_end->next;
			it_r_match_end->next->prev = it_r_match_begin->prev;
			it_r_match_begin->prev = NULL;
			it_r_match_end->next = NULL;	
		}
#endif
		//TODO: fix this mem free crash
#if 1
		//free the right portion
		cur = it_r_match_begin;
		end = it_r_match_end->next;
		while(cur != end && cur)
		{
			nxt = cur->next;
			delete cur;
			cur = nxt;	
		}
		//fprintf(STDOUT, "\nMerge LR Update");		
#endif
	}
	else
	{
		printf("\nERROR: intertask_mergeLeftRightPortions - both are not EPRSDs");
	}
	return;
}


	template <class T, template <class M> class MATCH, template <class G> class MERGE>
_INT64 EPRSD_COMPRESSOR<T, MATCH, MERGE>::mergeLeftRightPortions()
{
	EPRSD<T> *cur = NULL;
	EPRSD<T> *end = NULL;
	EPRSD<T> *right_cur = NULL;
	EPRSD<T> *nxt = NULL;	
	_INT64 stride = 0;
	_INT64 length = 0;
	_INT64 count = 0;
	_INT64 lcount = 0, rcount = 0;

	//fprintf(STDOUT, "\nmergeLeftRIghtPortions - entering");
	if(l_match_begin == NULL || l_match_end == NULL || r_match_begin == NULL || r_match_end == NULL)
	{
		fprintf(STDOUT, "\nError: mergeLeftRIghtPortions - NULL ptrs");
		l_match_begin = l_match_end = r_match_begin = r_match_end = NULL;
		return _ERROR;
	}

	if(l_match_begin->getAttr() & (ATTR_MEMBER|ATTR_TAIL))
	{
		//fprintf(STDOUT, "\nMEMBER/TAIL cannot become LEADER, merge ignored");
		l_match_begin = l_match_end = r_match_begin = r_match_end = NULL;
		return _ERROR;
	}

	cur = NULL;
	/* RSD or EPRSD is already formed, merge the newly found match*/
	if(l_match_begin->getAttr() & ATTR_LEADER && l_match_end->getAttr() & ATTR_TAIL)
	{
		//fprintf(STDOUT, "\nMerge LR EPRSD-EPRSD");
		//fflush(STDOUT);
		//update the length of RSDs
		right_cur = r_match_begin;
		rcount = 0;
		while(right_cur != r_match_end->next && right_cur != NULL)
		{
			rcount++;
			right_cur = right_cur->next;
		}
		cur = l_match_begin;
		lcount = 0;
		while(cur != l_match_end->next && cur != NULL)
		{
			lcount++;
			cur = cur->next;
		}
		//if(l_match_begin->getLoopSize() != count)
		if(lcount != rcount)
		{  
			fprintf(STDOUT, "\nError: Intratask - EPRSD merging - count mismatch, return w/o merging %d %d", 
					lcount, rcount);
			fprintf(STDOUT, "\nError:>>> Intratask - EPRSD merging - count mismatch, return w/o merging %d %d", 
					l_match_begin->getLoopSize(), r_match_begin->getLoopSize());
			l_match_begin = l_match_end = r_match_begin = r_match_end = NULL;
			return _ERROR;	
		}
		count = 0;
		cur = l_match_begin;
		right_cur = r_match_begin;
		while(cur != l_match_end->next)
		{
			/* increment the length of only outermost loop references */
			if(cur == l_match_begin)
			{
				//fprintf(STDOUT, "\n##%p##", cur);
				length = cur->getLength() + right_cur->getLength();
				//printf("$%d ", length);
				cur->setLength(length);
				if(cur && right_cur)
				{
					_UINT64 max_addr = 0, min_addr = 0;
					max_addr = _MAX(cur->userinfo->getValue(), right_cur->userinfo->getValue());			
					min_addr = _MIN(cur->userinfo->getValue(), right_cur->userinfo->getValue());			
					stride = (max_addr - min_addr)/(cur->getLength() - 1);
					//fprintf(STDOUT, "\nstride(h) %p = %p - %p", stride, (cur->userinfo)->getValue(), (right_cur->userinfo)->getValue());
					//fprintf(STDOUT, "\nstride(d) %ld = %ld - %ld", stride, (cur->userinfo)->getValue(), (right_cur->userinfo)->getValue());
				}
				cur->setStride(stride);
			}
			MERGE<T>::mergecb(cur->userinfo, right_cur->userinfo);
			cur = cur->next;

			if(right_cur)
				right_cur = right_cur->next;
			//compute the EPRSD length, # of RSDs
			count++;
		}

		//adjust the head/tail before deleting right portion
		l_match_end->next = r_match_end->next;
		if(r_match_end->next)
			r_match_end->next->prev = l_match_end;

		if(listtail == r_match_end)
			listtail = l_match_end;
		//free the right portion
		cur = r_match_begin;
		end = r_match_end->next;
		//fprintf(STDOUT, "\nMerge LR#1(b)");
		while(cur != end && cur)
		{
			nxt = cur->next;
			//fprintf(STDOUT, "\nfreeing %p, nxt = %p", cur, nxt);	
			delete cur;
			//fprintf(STDOUT, "\nfreed %p", cur);	
			cur = nxt;
		}
		//fprintf(STDOUT, "\nMerge LR#1 end");
	}
	else if(l_match_begin->getAttr() == ATTR_NONE && l_match_end->getAttr() == ATTR_NONE ||		
			l_match_begin->getAttr() == ATTR_NONE && l_match_end->getAttr() & ATTR_TAIL ||
			l_match_begin->getAttr() & ATTR_LEADER && l_match_end->getAttr() == ATTR_NONE
	       )
	{
		//fprintf(STDOUT, "\nMerge LR NONE-NONE");
		//fflush(STDOUT);
		addAfter(l_match_begin->createLoopHead(), l_match_begin->prev);
		l_match_begin->prev->setAttr(ATTR_LEADER); /* l-match_begin->prev is now the newly added node */
#ifdef SIGTREE_ENABLED
		l_match_begin->prev->cmninfo->sigtree_leaf = loophead_leaf;
#endif
		//update the length of RSDs
		cur = l_match_begin;
		right_cur = r_match_begin;
		count = 0;
		while(cur != l_match_end->next && right_cur != r_match_end->next && cur && right_cur)
		{
			_UINT64 max_addr = 0, min_addr = 0;
			//all nodes point to their parent EPRSD node
			if(cur == l_match_begin)
			{
				length = cur->getLength() + right_cur->getLength();
				//printf("$%d ", length);
				l_match_begin->prev->setLength(length);
				max_addr = _MAX(cur->userinfo->getValue(), right_cur->userinfo->getValue());
				min_addr = _MIN(cur->userinfo->getValue(), right_cur->userinfo->getValue());
				stride = (max_addr - min_addr)/(l_match_begin->prev->getLength() - 1);
				//fprintf(STDOUT, "\nstride(h) %p = %p - %p", stride, (cur->userinfo)->getValue(), (right_cur->userinfo)->getValue());
				//fprintf(STDOUT, "\nstride(d) %ld = %ld - %ld", stride, (cur->userinfo)->getValue(), (right_cur->userinfo)->getValue());
				l_match_begin->prev->setStride(stride);
			}
			MERGE<T>::mergecb(cur->userinfo, right_cur->userinfo);
			//cur->setCode(CODE_EPRSD);
			if(cur == l_match_end)
			{	cur->setAttr(ATTR_TAIL);	}
			else //if(cur != l_match_begin)
			{	cur->setAttr(ATTR_MEMBER);	}
			//compute the EPRSD length, # of RSDs			
			cur = cur->next;
			right_cur = right_cur->next;
			count++;
		}
		//l_match_begin->setLoopSize(count);
		l_match_begin->prev->setLoopSize(count+1);

		//adjust the head/tail before deleting right portion
		l_match_end->next = r_match_end->next;
		if(r_match_end->next)
			r_match_end->next->prev = l_match_end;
		if(listtail == r_match_end)
			listtail = l_match_end;

		//free the right portion
		cur = r_match_begin;
		end = r_match_end->next;
		while(cur && cur != end)
		{
			nxt = cur->next;
			delete cur;
			cur = nxt;
		}

		//latest_eprsd = l_match_begin;
		//fprintf(STDOUT, "\nMerge LR end - %p", latest_eprsd);
	}
#if 0
	else if(l_match_begin->getAttr() & ATTR_LEADER && l_match_end->getAttr() == ATTR_NONE)
	{
		fprintf(STDOUT, "\nError: Head of two loops case");
		//l_match_begin = l_match_end = r_match_begin = r_match_end = NULL;
		//return _ERROR;		
		addAfter(l_match_begin->createLoopHead(), l_match_begin->prev);
		l_match_begin->prev->setAttr(ATTR_LEADER); /* l-match_begin->prev is now the newly added node */
		//update the length of RSDs
		cur = l_match_begin;
		right_cur = r_match_begin;
		count = 0;
		while(cur != l_match_end->next && right_cur != r_match_end->next && cur && right_cur)
		{
			_UINT64 max_addr = 0, min_addr = 0;
			//all nodes point to their parent EPRSD node
			if(cur == l_match_begin)
			{
				length = cur->getLength() + right_cur->getLength();			
				//printf("$%d ", length);
				l_match_begin->prev->setLength(length);
				max_addr = _MAX(cur->userinfo->getValue(), right_cur->userinfo->getValue());			
				min_addr = _MIN(cur->userinfo->getValue(), right_cur->userinfo->getValue());			
				stride = (max_addr - min_addr)/(l_match_begin->prev->getLength() - 1);
				//fprintf(STDOUT, "\nstride(h) %p = %p - %p", stride, (cur->userinfo)->getValue(), (right_cur->userinfo)->getValue());
				//fprintf(STDOUT, "\nstride(d) %ld = %ld - %ld", stride, (cur->userinfo)->getValue(), (right_cur->userinfo)->getValue());
				l_match_begin->prev->setStride(stride);
			}
			MERGE<T>::mergecb(cur->userinfo, right_cur->userinfo);
			//cur->setCode(CODE_EPRSD);
			if(cur == l_match_end)
			{	cur->setAttr(ATTR_TAIL);	}
			else //if(cur != l_match_begin)
			{	cur->setAttr(ATTR_MEMBER);	}
			//compute the EPRSD length, # of RSDs			
			cur = cur->next;
			right_cur = right_cur->next;
			count++;						
		}
		//l_match_begin->setLoopSize(count);
		l_match_begin->prev->setLoopSize(count+1);

		//adjust the head/tail before deleting right portion
		l_match_end->next = r_match_end->next;
		if(r_match_end->next)
			r_match_end->next->prev = l_match_end;
		if(listtail == r_match_end)
			listtail = l_match_end;

		//free the right portion
		cur = r_match_begin;
		end = r_match_end->next;
		while(cur && cur != end)
		{
			nxt = cur->next;
			delete cur;
			cur = nxt;
		}
	}
#endif
#if 0
	else if(l_match_begin->getAttr() == ATTR_NONE && l_match_end->getAttr() & ATTR_TAIL)//first sequence match
	{
		fprintf(STDOUT, "\nError: Merge LR NONE-TAIL");
		l_match_begin = l_match_end = r_match_begin = r_match_end = NULL;
		return _ERROR;
	}
#endif
	else
	{
		fprintf(STDOUT, "\nError: unknown case, merge ignored");
		l_match_begin = l_match_end = r_match_begin = r_match_end = NULL;
		return _ERROR;
	}

	l_match_begin = l_match_end = r_match_begin = r_match_end = NULL;
	return _SUCCESS;
}

	template <class T>
EPRSD<T>* EPRSD<T>::createLoopHead()
{
	_UINT64 signature[1]={LOOP_HEAD_SIGN};
	COMMON* cmn = new COMMON(MAX_SIGN_LEN);
	T* address = new T();	
	cmn->setData(signature, 1, signature[0], 'h', 0, 0, "LH");	
	EPRSD<T>* neweprsd = new EPRSD<T>(cmn, address);
	neweprsd->setAttr(ATTR_LEADER);
	return neweprsd;
}

	template <class T>
EPRSD<T>* EPRSD<T>::createLoopTail()
{
	COMMON* cmn = new COMMON(MAX_SIGN_LEN);
	T* address = new T();
	EPRSD<T>* neweprsd = new EPRSD<T>(cmn, address);
	neweprsd->setAttr(ATTR_TAIL);
	return neweprsd;
}

/* add trace info to compressor */
	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::add(COMMON* cmn, T* newentry)
{
	EPRSD<T> *match = NULL;
	EPRSD<T> *neweprsd = NULL;
#ifdef IMPRECISION_ENABLED
	_UINT64 n_match = 0;
	_UINT64 n_mismatch = 0;
#endif	
	bool repeat = false;
	static int growing = 0;	
	try
	{
		neweprsd = new EPRSD<T>(cmn, newentry);
	}
	catch(bad_alloc &e)
	{
		fprintf(STDOUT, "\nadd() new() failed");
		return;
	}
	//fprintf(STDOUT, "\n*****************add starts*****************");

#ifdef SIGTREE_ENABLED
	if(!sigtree){fprintf(STDOUT, "\nError: SIGTREE NULL pointer");}
	void* pc_ra = NULL;
	_INT32 i = 0, sign_len = 0;
	//indicate the tree object that adding pc/ra is going to start
	sigtree->initAdding();
	sign_len = neweprsd->cmninfo->sign_len;
	//printf("\nsign arr: ");
	for(i = 0; i < sign_len; i++)
	{
		pc_ra = (void*)neweprsd->cmninfo->signature[i];
		neweprsd->cmninfo->sigtree_leaf = (SIGNODE*)sigtree->add(pc_ra); /* pointer to signature tree leaf node */
		//printf("%p, ", pc_ra);		
	}	
	sigtree->finishAdding();
#endif

	/* add at the tail end */
	addTail(neweprsd);
	do
	{
#ifdef IMPRECISION_ENABLED
		n_match = 0;
		n_mismatch = 0;
#endif	
		repeat = false;
		match = NULL;	
		r_match_begin = NULL;
		r_match_end = NULL;
		l_match_begin = NULL;
		l_match_end = NULL;
		match = findTailMatch();	
		if(match)
		{
			//fprintf(STDOUT, "\nmatch - %p", (void*)match->cmninfo->ref_id);
			r_match_begin = listtail;
			r_match_end = listtail;
			l_match_begin = match;
			l_match_end = match;

			/* search for matching sequence */
			while(l_match_end != r_match_begin->prev)
			{
				if(l_match_begin->prev && r_match_begin->prev)
				{
#ifdef SIGTREE_ENABLED
					if( 	l_match_begin->prev->cmninfo->sigtree_leaf == r_match_begin->prev->cmninfo->sigtree_leaf
							&& MATCH<T>::compare(l_match_begin->prev->userinfo, r_match_begin->prev->userinfo)
							&& l_match_begin->prev->getAttr() ==  r_match_begin->prev->getAttr()
							//&& l_match_begin->prev->getLength() == r_match_begin->prev->getLength() 
					  )
#else
						if( 	isSignMatch(l_match_begin->prev->cmninfo, r_match_begin->prev->cmninfo)
								&& MATCH<T>::compare(l_match_begin->prev->userinfo, r_match_begin->prev->userinfo)
								&& l_match_begin->prev->getAttr() ==  r_match_begin->prev->getAttr()
								//&& l_match_begin->prev->getLength() == r_match_begin->prev->getLength() 
						  )
#endif
						{
							//fprintf(STDOUT, "\nmatching continues");
							//fprintf(STDOUT, " %p %p\n",(void*)l_match_begin->prev->cmninfo->ref_id, (void*)r_match_begin->prev->cmninfo->ref_id);
							l_match_begin = l_match_begin->prev;
							r_match_begin = r_match_begin->prev;
#ifdef IMPRECISION_ENABLED
							n_match = n_match + l_match_begin->getLength();
#endif
						}
						else /* matching stops */
						{
#ifdef IMPRECISION_ENABLED
							//fprintf(STDOUT, "\nmatching stops, but imprecision allowed...continue matching");
							//fprintf(STDOUT, " %p %p\n",(void*)l_match_begin->prev->cmninfo->ref_id, (void*)r_match_begin->prev->cmninfo->ref_id);
							l_match_begin = l_match_begin->prev;
							r_match_begin = r_match_begin->prev;
							n_mismatch = n_mismatch + l_match_begin->getLength();
#else
							//fprintf(STDOUT, "\nmatching stops, sign/depth mismatch");
							//fprintf(STDOUT, " %p %p\n",(void*)l_match_begin->prev->cmninfo->ref_id, (void*)r_match_begin->prev->cmninfo->ref_id);
							break;
#endif
						}
				}
				else /* beginning of list reached */
				{
					//fprintf(STDOUT, "\nmatching stops, front of list reached");				
					break;
				}
			}

#ifdef IMPRECISION_ENABLED
			float mismatch_percentage = 0.0;
			mismatch_percentage = (100.0*n_mismatch)/(n_match+n_mismatch);
			//fprintf(STDOUT, "\nPercentage of mismatch = %.2f",mismatch_percentage);
#endif

			if(l_match_begin && l_match_end && r_match_begin && r_match_end)
			{		
				if(l_match_end == r_match_begin->prev) /* matching sequence found, merge :-) */
				{
					//fprintf(STDOUT, "\nmatching sequence found L<%p %p> R<%p %p>", (void*)l_match_begin->cmninfo->ref_id, (void*)l_match_end->cmninfo->ref_id, (void*)r_match_begin->cmninfo->ref_id, (void*)r_match_end->cmninfo->ref_id);
					//fprintf(STDOUT, "\nmatching sequence ptr L<%p %p> R<%p %p>", (void*)l_match_begin, (void*)l_match_end, (void*)r_match_begin, (void*)r_match_end);
#ifdef IMPRECISION_ENABLED
					if(mismatch_percentage <= 10.0)
					{
						//fprintf(STDOUT, "\nmerging with imprecision < 10\%");	
						if(mergeLeftRightPortions() == _SUCCESS)
						{
							repeat = true;
							//fprintf(STDOUT, "\nmerge SUCCESSful");
						}
						else
						{				
							repeat = false;
							//fprintf(STDOUT, "\nmerge ERROR");					
						}	
					}
					else
					{
						repeat = false;
					}
#else
					if(mergeLeftRightPortions() == _SUCCESS)
					{
						//fprintf(STDOUT, "\nmerge SUCCESS");
						repeat = true;
					}
					else
					{
						//fprintf(STDOUT, "\nmerge ERROR");
						repeat = false;
					}	
#endif
				}
				else if(l_match_begin->prev == NULL)
				{
					//fprintf(STDOUT, "\nfront of list reached 2");
				}
				else
				{
					//fprintf(STDOUT, "\nsequence broken in b/w due to sign/depth mismatch");
				}
			}
			else
			{
				//fprintf(STDOUT, "\nNULL window ptrs - front of list reached");
			}
		}/* if(match) ends */
		else /* no match found, remains added at tail end */
		{
			//fprintf(STDOUT, "\nmatch NOT found in the whole list");
			repeat = true;
		}
		//getchar();
	}while(match && repeat);

	r_match_begin = NULL;
	r_match_end = NULL;
	l_match_begin = NULL;
	l_match_end = NULL;
	//fprintf(STDOUT, "\n*****************add ends*****************");
	return;
}


	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::internode_merge(EPRSD<T>* match, EPRSD<T>* neweprsd)
{
	int length = 1, stride = 0, min_stride = 0, max_stride = 0;
	length = match->getNodeIdLength() + neweprsd->getNodeIdLength();
	match->setNodeIdLength(length);
	match->setNodeIdBits(neweprsd->getNodeIdBits());
	min_stride = _MIN(match->getNodeId(), neweprsd->getNodeId());
	max_stride = _MAX(match->getNodeId(), neweprsd->getNodeId());
	stride = max_stride - min_stride;
	//stride = (_abs( (match->getNodeId()) - (neweprsd->getNodeId()) )); // /((match->getThreadIdLength()) - 1);
	match->setNodeIdStride(stride);
	return;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::interthread_merge(EPRSD<T>* match, EPRSD<T>* neweprsd)
{
	_INT64 length = 1, stride = 0;
	_INT64 max_tid = 0, min_tid = 0, max_addr = 0, min_addr = 0;
	//printf("\nintertask_merge - match tid: %ld, neweprsd tid: %ld", match->getThreadId(), neweprsd->getThreadId());
	length = match->getThreadIdLength() + neweprsd->getThreadIdLength();
	match->setThreadIdLength(length);
	match->setThreadIdBits(neweprsd->getThreadIdBits());
	min_tid = _MIN(match->getThreadId(), neweprsd->getThreadId());
	max_tid = _MAX(match->getThreadId(), neweprsd->getThreadId());
	//match->setThreadId(min_tid);
	stride = max_tid - min_tid;
	//stride = (_abs( (match->getThreadId()) - (neweprsd->getThreadId()) )); // /((match->getThreadIdLength()) - 1);
	match->setThreadIdStride(stride);
	/* address as a function of tid */
	max_addr = _MAX(match->getStartValue(), neweprsd->getStartValue());
	min_addr = _MIN(match->getStartValue(), neweprsd->getStartValue());
	if(max_tid - min_tid)
	{	
		stride = (max_addr - min_addr)/(max_tid - min_tid);	
		match->setThreadIdAddrStride(stride);
	}
	return;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::internode_startMerge()
{
	startMerge(INTERNODE_MERGE);
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::interthread_startMerge()
{
	startMerge(INTERTHREAD_MERGE);
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::startMerge(_INT64 merge_type)
{
	EPRSD<T> *match = NULL;
	EPRSD<T> *rcur = NULL;
	EPRSD<T> *nxt = NULL;
	const char* type[] = {"IT", "IN"};
	static int matched_count = 0;
	static int not_matched_count = 0;
	static int merged_none_count = 0;
	static int not_merged_none_count = 0;
	static int merged_loop_count = 0;
	static int not_merged_loop_count = 0;
	int match_broken = 0;
	static int counter = 0;

	fprintf(STDOUT, "%s(%d): starting merge\n", type[merge_type], rank);
	l_match_begin = l_match_end = r_match_begin = r_match_end = NULL;
	prev_merge_point = NULL;
	matched_count = 0;  not_matched_count = 0; merged_none_count = 0; not_merged_none_count = 0;
	merged_loop_count = 0; not_merged_loop_count = 0;

	rcur = rlisthead; //src list
	counter = 0;
	while(rcur)
	{	
		if(merge_type == INTERNODE_MERGE)
			match = internode_findExactMatch(rcur);
		else if(merge_type == INTERTHREAD_MERGE)
			match = interthread_findExactMatch(rcur);

		if(match)/* case 1: independent node, case 2: loop */
		{
			matched_count++;
			if(match->getAttr() == ATTR_NONE)
			{	
				nxt = rcur->next;
				if(merge_type == INTERNODE_MERGE)
					internode_merge(match, rcur);
				else if(merge_type == INTERTHREAD_MERGE)
					interthread_merge(match, rcur);

				removeNoder(rcur);
				merged_none_count++;				
				prev_merge_point = match; //IMPORTANT
			}
			else if(match->getAttr() & ATTR_LEADER)
			{
				EPRSD<T>* lprevious = NULL;
				EPRSD<T>* rprevious = NULL;
				EPRSD<T>* l_seq_end = NULL;
				EPRSD<T>* r_seq_end = NULL;
				//fprintf(STDOUT, "\n%s(%d): found loop sequence", type[merge_type], rank);
				l_match_begin = match;
				l_match_end = match;
				r_match_begin = rcur;
				r_match_end = rcur;
				match_broken = 0;

				/* compare node to node in the two sequences - all must match */
				while(l_match_end->getAttr() != ATTR_NONE && r_match_end->getAttr() != ATTR_NONE)
				{
					//if nodes in the two sequences match, l_match_end and r_match_end point to the respective tail ends	
					if(merge_type == INTERNODE_MERGE)
					{
#ifdef SIGTREE_ENABLED
						if( l_match_end->cmninfo->sigtree_leaf == r_match_end->cmninfo->sigtree_leaf && 
								MATCH<T>::internode_compare(l_match_end->userinfo, r_match_end->userinfo) &&
								l_match_end->getAttr() == r_match_end->getAttr() &&
								(l_match_end->getLoopSize() == r_match_end->getLoopSize()) &&
								(l_match_end->getLoopCount() == r_match_end->getLoopCount()) &&
								(l_match_end->getStride() == r_match_end->getStride())&&
								(l_match_end->getThreadIdLength() == r_match_end->getThreadIdLength())&&
								(l_match_end->getThreadIdStride() == r_match_end->getThreadIdStride())&&
								(l_match_end->getThreadId() == r_match_end->getThreadId()) &&
								(l_match_end->getNodeIdLength() == r_match_end->getNodeIdLength())&&
								(l_match_end->getNodeIdStride() == r_match_end->getNodeIdStride())
						  ) 
#else
							if( isSignMatch(l_match_end->cmninfo, r_match_end->cmninfo)&& 
									MATCH<T>::internode_compare(l_match_end->userinfo, r_match_end->userinfo) &&
									l_match_end->getAttr() == r_match_end->getAttr() &&
									(l_match_end->getLoopSize() == r_match_end->getLoopSize()) &&
									(l_match_end->getLoopCount() == r_match_end->getLoopCount()) &&
									(l_match_end->getStride() == r_match_end->getStride())&&
									(l_match_end->getThreadIdLength() == r_match_end->getThreadIdLength())&&
									(l_match_end->getThreadIdStride() == r_match_end->getThreadIdStride())&&
									(l_match_end->getThreadId() == r_match_end->getThreadId()) &&
									(l_match_end->getNodeIdLength() == r_match_end->getNodeIdLength())&&
									(l_match_end->getNodeIdStride() == r_match_end->getNodeIdStride())
							  ) 
#endif
							{
								;
							}
							else
							{
								//fprintf(STDOUT, "\nIN: loop sequence match broken");
								match_broken = 1;
							}
					}
					else if(merge_type == INTERTHREAD_MERGE)
					{
#ifdef SIGTREE_ENABLED
						if( l_match_end->cmninfo->sigtree_leaf == r_match_end->cmninfo->sigtree_leaf && 
								MATCH<T>::interthread_compare(l_match_end->userinfo, r_match_end->userinfo) &&
								l_match_end->getAttr() == r_match_end->getAttr() &&
								(l_match_end->getLoopSize() == r_match_end->getLoopSize()) &&
								(l_match_end->getLoopCount() == r_match_end->getLoopCount()) &&
								(l_match_end->getStride() == r_match_end->getStride())
						  ) 
#else
							if( isSignMatch(l_match_end->cmninfo, r_match_end->cmninfo)&& 
									MATCH<T>::interthread_compare(l_match_end->userinfo, r_match_end->userinfo) &&
									l_match_end->getAttr() == r_match_end->getAttr() &&
									(l_match_end->getLoopSize() == r_match_end->getLoopSize()) &&
									(l_match_end->getLoopCount() == r_match_end->getLoopCount()) &&
									(l_match_end->getStride() == r_match_end->getStride())
							  ) 
#endif
							{
								;
							}
							else
							{
								//fprintf(STDOUT, "\nIT: loop sequence match broken");
								match_broken = 1;
							}
					}
					lprevious = l_match_end;
					rprevious = r_match_end;
					l_match_end = l_match_end->next;
					r_match_end = r_match_end->next;

					if(!l_match_end || !r_match_end)
						break;
				}//while ends

				/* both sequences may reach end together or not - so check all possible conditions */
				if(l_match_end && r_match_end)
				{	
					if(l_match_end->getAttr() == ATTR_NONE && r_match_end->getAttr() == ATTR_NONE)
					{
						//fprintf(STDOUT, "\n%s(%d): both have reached ATTR_NONE simultaneously #1",type[merge_type], rank);
					}				
					/* move to end of loop */
					while(l_match_end->getAttr() != ATTR_NONE)
						l_match_end = l_match_end->next;
					while(r_match_end->getAttr() != ATTR_NONE)
						r_match_end = r_match_end->next;				
					if(l_match_end->getAttr() == ATTR_NONE && r_match_end->getAttr() == ATTR_NONE)
					{
						//fprintf(STDOUT, "\n%s(%d): both have reached ATTR_NONE simultaneously #2",type[merge_type], rank);
					}
					l_seq_end = l_match_end->prev;
					r_seq_end = r_match_end->prev;	
				}
				else
				{
					match_broken = 1;
					r_seq_end = rprevious;
					//fprintf(STDOUT, "\n%s(%d): end of loop was end of list for one seq L(%p) R(%p)",type[merge_type], rank, l_match_end, r_match_end);
					if(r_match_end)
					{
						//fprintf(STDOUT, "\n%s(%d): src list reached end - adjusting r_match_end", type[merge_type], rank);
						while(r_match_end->getAttr() != ATTR_NONE)
						{
							rprevious = r_match_end;
							r_match_end = r_match_end->next;
							if(!r_match_end)
								break;
						}
						if(r_match_end == NULL)
							r_seq_end = rprevious;
					}
				}
#if 0
				else if(l_match_end && r_match_end == NULL)
				{
					if(l_match_end->getAttr() == ATTR_NONE)
					{
						fprintf(STDOUT, "\nIN(%d): L ATTR_NONE, R NULL simultaneously #3", rank);
						l_seq_end = l_match_end->prev;
						r_seq_end = rprevious;
					}
					else
					{
						match_broken = 1;
						r_seq_end = rprevious;
						fprintf(STDOUT, "\nIN(%d): L !ATTR_NONE, R NULL simultaneously #3", rank);
					}					
				}
				else if(l_match_end == NULL && r_match_end)
				{

				}
				else if(l_match_end == NULL && r_match_end == NULL)
				{

				}
#endif								

				if(!match_broken)//merge the sequence
				{	
					EPRSD<T>* l = NULL;
					EPRSD<T>* r = NULL;
					EPRSD<T>* n = NULL;
					//fprintf(STDOUT, "\n%s(%d): loop sequence match",type[merge_type], rank);
					merged_loop_count++;
					prev_merge_point = l_match_end->prev;							
					nxt = r_match_end;
					/* merge the nodes in sequence */
					l = l_match_begin;
					r = r_match_begin;
					while(l != l_match_end && r != r_match_end)
					{
						if(merge_type == INTERNODE_MERGE)
							internode_merge(l, r);
						else if(merge_type == INTERTHREAD_MERGE)
							interthread_merge(l, r);
						l = l->next;
						r = r->next;
					}
					if(l->getAttr() == ATTR_NONE && r->getAttr() == ATTR_NONE)
					{							
						//fprintf(STDOUT, "\n%s(%d): Merge Successful, both have reached ATTR_NONE simultaneously #3",type[merge_type], rank);
					}
					/* free the sequence in src list */
					r = r_match_begin;
					while(r != r_match_end)
					{
						n = r->next;
						removeNoder(r);
						r = n;
					}
				}
				else
				{
					EPRSD<T>* start = NULL;
					EPRSD<T>* end = NULL;
					//fprintf(STDOUT, "\n%s(%d): loop sequence PARTIAL match", type[merge_type], rank);
					/* src sequence added before dest sequence, if match not found */
					//before merging do this
					//in_prev_merge_point = r_match_end->prev;
					//in_prev_merge_point = l_match_end->prev;
					nxt = r_match_end;
					/* move the sequence from src to dest */
					start = r_match_begin;
					if(r_match_end)
						end = r_match_end->prev;
					else
						end = r_seq_end;
					if(start->prev)
						start->prev->next = end->next;
					else
					{	rlisthead = end->next; if(rlisthead) rlisthead->prev = NULL; }
					if(end->next)
						end->next->prev = start->prev;
					else
					{	rlisttail = start->prev; if(rlisttail) rlisttail->next = NULL;}

					start->prev = NULL;
					end->next = NULL;
					insertSequenceAfter(prev_merge_point, start, end);
					prev_merge_point = end;
					not_merged_loop_count++;
				}
			}
			else
			{
				fprintf(STDOUT, "\nError: %s(%d): startMerge() - isolated MEMBER/TAIL found attr: %d", type[merge_type], rank, match->getAttr());
				nxt = rcur->next;
			}
		}//if match ends
		else//if match not found, simply add the node to the dest list after last merge point
		{

			not_matched_count++;
			if(rcur->getAttr() == ATTR_NONE)
			{
				EPRSD<T>* start = NULL;
				EPRSD<T>* end = NULL;
				//fprintf(STDOUT, "\n%s(%d):node DO NOT match", type[merge_type], rank);
				nxt = rcur->next;
				start = rcur;
				end = rcur;
				if(start->prev)
					start->prev->next = end->next;
				else
				{	rlisthead = end->next; if(rlisthead) rlisthead->prev = NULL; }
				if(end->next)
					end->next->prev = start->prev;
				else
				{	rlisttail = start->prev; if(rlisttail) rlisttail->next = NULL; }

				start->prev = NULL;
				end->next = NULL;
				insertSequenceAfter(prev_merge_point, start, end);
				prev_merge_point = end;
				not_merged_none_count++;
			}
			else if(rcur->getAttr() & ATTR_LEADER)
			{
				EPRSD<T>* start = NULL;
				EPRSD<T>* end = NULL;
				EPRSD<T>* previous = NULL;
				//fprintf(STDOUT, "\n%s(%d):loop NOT found, DO NOT match", type[merge_type], rank);	fflush(STDOUT);
				r_match_begin = rcur;
				r_match_end = rcur;
				while(r_match_end->getAttr() != ATTR_NONE)
				{
					previous = r_match_end;
					r_match_end = r_match_end->next;
					if(!r_match_end)
					{
						//fprintf(STDOUT, "\n%s(%d): loop is also the end of list", type[merge_type], rank); fflush(STDOUT);
						break;
					}
				}
				if(r_match_end)
				{
					//fprintf(STDOUT, "\n%s(%d): reached ATTR_NONE #4", type[merge_type], rank); fflush(STDOUT);
					end = r_match_end->prev;
				}
				else
				{
					//fprintf(STDOUT, "\n%s(%d): NOT reached ATTR_NONE #5 - loop end of list", type[merge_type], rank); fflush(STDOUT);
					end = previous;					
				}
				start = r_match_begin;				
				if(start->prev)
					start->prev->next = end->next;
				else
				{	rlisthead = end->next; if(rlisthead) rlisthead->prev = NULL; }
				if(end->next)
					end->next->prev = start->prev;
				else
				{	rlisttail = start->prev; if(rlisttail) rlisttail->next = NULL;}

				start->prev = NULL;
				end->next = NULL;
				//fprintf(STDOUT, "\n%s(%d): seq separated from src", type[merge_type], rank); fflush(STDOUT);
				insertSequenceAfter(prev_merge_point, start, end);
				prev_merge_point = end;
				not_merged_loop_count++;	
				nxt = r_match_end;
			}
			else
			{
				fprintf(STDOUT, "\nError: %s(%d): startMerge() - isolated MEMBER/TAIL found #2 attr: %d", type[merge_type], rank, match->getAttr());
				nxt = rcur->next;
			}
		}
		rcur = nxt;
		counter++;
		if(counter%1000 == 0)
			fprintf(STDOUT, "%d\r", counter);
	}
	fprintf(STDOUT, "\n%s(%d): matched %d", type[merge_type],  rank, matched_count);
	fprintf(STDOUT, "\n%s(%d): not_matched %d",type[merge_type], rank, not_matched_count);
	fprintf(STDOUT, "\n%s(%d): merged_none %d",type[merge_type], rank, merged_none_count);
	fprintf(STDOUT, "\n%s(%d): not_merged_none_count %d",type[merge_type], rank, not_merged_none_count);
	fprintf(STDOUT, "\n%s(%d): merged_loop_count %d",type[merge_type], rank, merged_loop_count);
	fprintf(STDOUT, "\n%s(%d): not_merged_loop_count %d",type[merge_type], rank, not_merged_loop_count);
	return;
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::internode_add(EPRSD_DATA<T> *data)
{
	static int growing = 0;	
	//if it is an RSD, just merge
	if(!data)
		return;

	EPRSD<T> *neweprsd = new EPRSD<T>(data);	

	/* convert virtual signature to real signature - use key to find and associate the leaf node to this new EPRSD*/
#ifdef SIGTREE_ENABLED
	if(neweprsd->cmninfo->sign_type == VIRTUAL)
	{
		SIGDATA newdata;
		SIGNODE *leaf = NULL;
		bool ret = false;
		newdata.info = (void*)neweprsd->cmninfo->signature[0];
		newdata.key[0] = neweprsd->cmninfo->signature[1]; 	
		newdata.key[1] = neweprsd->cmninfo->signature[2]; 	
		newdata.key[2] = neweprsd->cmninfo->signature[3];
		sigtree->getLeafNode(&newdata, &leaf);

		if(!leaf)
		{
			printf("\nError: IN: Leaf node not found for new EPRSD from another thread");
		}
		else
		{
			printf("\nSuccess: IN: Leaf node found for new EPRSD from another thread");
		} 	
		neweprsd->cmninfo->sigtree_leaf = leaf;
		/* store the actual signature in place of virtual signature */
		neweprsd->cmninfo->sign_len = neweprsd->cmninfo->max_sign_len;
		/* below function returns the actual signature length at the location whose address is contained in the third argument. Maximum signature
		   array size is input in the third argument too. */
		ret = sigtree->getSignatureFromLeaf(leaf, (void**)neweprsd->cmninfo->signature, &(neweprsd->cmninfo->sign_len));
		if(!ret)	
		{
			printf("\nError: IN: obtaining signature from leaf node");
		}	
	}	
#endif

	addTailr(neweprsd);

#if 0	
	while(1)
	{
		//fprintf(STDOUT, "\nIN: neweprsd: %p len: %ld", (void*)neweprsd->cmninfo->ref_id, neweprsd->getLength()); 
		match = internode_findExactMatch(neweprsd);
		if(match)	
		{
			//fprintf(STDOUT, "\nInter node - match found neweprsd: %p match: %p", (void*)neweprsd->cmninfo->ref_id, (void*)match->cmninfo->ref_id);
			if(match->getAttr() & ATTR_LEADER || ((match->getAttr() & (ATTR_MEMBER|ATTR_TAIL)) && growing))/* immediately merge the RSDs*/
			{					
				//internode_merge(match, neweprsd);
				//delete neweprsd; /* neweprsd not needed after merging */
#if 1
				addTail(neweprsd);
				if(in_l_match_begin == NULL)/* first item of EPRSD sequence */
				{
					in_l_match_begin = match;
					in_l_match_end = match;
					in_r_match_begin = neweprsd;
					in_r_match_end = neweprsd;
					growing++;
				}
				else /* subsequent items of EPRSD sequence */
				{
					/* matching continues */
					if(match == in_l_match_end->next)
					{
						//fprintf(STDOUT, "\nintertask window growing...len: %ld", match->getLength());
						//fprintf(STDOUT, "\ngrowing %p %p", (void*)it_l_match_end->cmninfo->ref_id, (void*)it_r_match_end->cmninfo->ref_id);
						in_l_match_end = match;
						in_r_match_end = neweprsd;
						growing++;
					}
					else /* matching stops, reset window pointers */
					{
						float percent = 0;
						fprintf(STDOUT, "\ninternode window stops...growing = %d  of %d", growing, in_l_match_begin->getLoopSize());
#if 0
						/* if it is 80% match, then merge */
						percent = growing*100.0/(in_l_match_begin->getLoopSize());
						if(percent > PERCENT_MATCH)
						{
							//fprintf(STDOUT, "\nintertask EPRSD merging - partial match");
							internode_mergeLeftRightPortions();							
							in_prev_merge_point = in_l_match_end;
						}
						else
						{
							/* include the new entry added as tail in the sequence */
							in_r_match_end = neweprsd;
							/* add sequence after in_prev_merge_point */
							internode_addSequenceAfter(in_prev_merge_point);
							in_prev_merge_point = neweprsd;
						}
#endif	
						in_l_match_begin = NULL;
						in_r_match_begin = NULL;
						in_l_match_end = NULL;
						in_r_match_end = NULL;
						growing = 0;
					}
					/* merge the EPRSD sequence */
					/* EPRSD sequence completed */
					if(growing)
					{
						if(in_l_match_begin->getLoopSize() == growing)
						{
							fprintf(STDOUT, "\n internode EPRSD merging - whole seq matched");
							internode_mergeLeftRightPortions();
							in_prev_merge_point = in_l_match_end;
							/* reset the pointers as they are no more required */
							in_l_match_begin = NULL;
							in_r_match_begin = NULL;
							in_l_match_end = NULL;
							in_r_match_end = NULL;
							growing = 0;
						}
					}
				}//else ends
			}
			else if(match->getAttr() & (ATTR_MEMBER|ATTR_TAIL))/* in case growing stopped, put it next to its predecessors*/
			{
				if(in_l_match_begin || in_l_match_end || in_r_match_begin || in_r_match_end)
					fprintf(STDOUT, "\nERROR: IN: L & R not NULL#1, growing = %d", growing);
				//addTail(neweprsd);
				addAfter(neweprsd, in_prev_merge_point);
				in_prev_merge_point = neweprsd;
			}
			else if(match->getAttr() == ATTR_NONE)/* immediately merge the RSDs*/
			{	
				if(in_l_match_begin || in_l_match_end || in_r_match_begin || in_r_match_end)
					fprintf(STDOUT, "\nERROR: IN: L & R not NULL#2, growing = %d", growing);
				internode_merge(match, neweprsd);
				in_prev_merge_point = match;
				delete neweprsd; /* neweprsd not needed after merging */
			}
			else
			{
				fprintf(STDOUT, "ERROR: IN: Invalid attr %d.....", match->getAttr());	
			}
#endif
		}//if(match) ends
		else
		{
			//addTail(neweprsd);
			addAfter(neweprsd, in_prev_merge_point);
			in_prev_merge_point = neweprsd;				
		}
		break;
	}//while ends
#endif

	//if an EPRSD, add as tail and look for the complete sequence match
	//when whole sequence is matched, merge the EPRSD
}

	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::interthread_add(EPRSD_DATA<T> *data)
{
	EPRSD<T> *match = NULL;
	static int growing = 0;	
	//if it is an RSD, just merge
	if(!data)
		return;

	EPRSD<T> *neweprsd = new EPRSD<T>(data);

	/* convert virtual signature to real signature - use key to find and associate the leaf node to this new EPRSD*/
#ifdef SIGTREE_ENABLED
	if(neweprsd->cmninfo->sign_type == VIRTUAL)
	{
		SIGDATA newdata;
		SIGNODE *leaf = NULL;
		bool ret = false;
		newdata.info = (void*)neweprsd->cmninfo->signature[0];
		newdata.key[0] = neweprsd->cmninfo->signature[1]; 	
		newdata.key[1] = neweprsd->cmninfo->signature[2]; 	
		newdata.key[2] = neweprsd->cmninfo->signature[3];
		sigtree->getLeafNode(&newdata, &leaf);

		if(!leaf)
		{
			printf("\nError: Leaf node not found for new EPRSD from another thread");
		}
		else
		{
			printf("\nSuccess: Leaf node found for new EPRSD from another thread");
		} 	
		neweprsd->cmninfo->sigtree_leaf = leaf;
		/* store the actual signature in place of virtual signature */
		neweprsd->cmninfo->sign_len = neweprsd->cmninfo->max_sign_len;
		/* below function returns the actual signature length at the location whose address is contained in the third argument. Maximum signature
		   array size is input in the third argument too. */
		ret = sigtree->getSignatureFromLeaf(leaf, (void**)neweprsd->cmninfo->signature, &(neweprsd->cmninfo->sign_len));
		if(!ret)	
		{
			printf("\nError: obtaining signature from leaf node");
		}	
	}	
#endif

	addTailr(neweprsd);
#if 0
	while(1)
	{
		//fprintf(STDOUT, "\nneweprsd: %p len: %ld", (void*)neweprsd->cmninfo->ref_id, neweprsd->getLength()); 
		match = intertask_findExactMatch(neweprsd);
		if(match)	
		{
#if 0 /* plain RSD match without looking for whole sequence*/
			intertask_merge(match, neweprsd);
			delete neweprsd; /* neweprsd not needed after merging */
#else/* looking for whole sequence match */			
			//fprintf(STDOUT, "\nInter task - match found neweprsd: %p match: %p", (void*)neweprsd->cmninfo->ref_id, (void*)match->cmninfo->ref_id);
			if(match->getAttr() & ATTR_LEADER || ((match->getAttr() & (ATTR_MEMBER|ATTR_TAIL)) && growing) )/* add to tail and wait for complete match sequence by building window*/
			{
				//intertask_merge(match, neweprsd);								
				//delete neweprsd; /* neweprsd not needed after merging */
#if 1
				addTail(neweprsd);
				if(it_l_match_begin == NULL)/* first item of EPRSD sequence */
				{
					it_l_match_begin = match;
					it_l_match_end = match;
					it_r_match_begin = neweprsd;
					it_r_match_end = neweprsd;
					growing++;
				}
				else /* subsequent items of EPRSD sequence */
				{
					/* matching continues */
					if(match == it_l_match_end->next)
					{
						//fprintf(STDOUT, "\nintertask window growing...len: %ld", match->getLength());
						//fprintf(STDOUT, "\ngrowing %p %p", (void*)it_l_match_end->cmninfo->ref_id, (void*)it_r_match_end->cmninfo->ref_id);
						it_l_match_end = match;
						it_r_match_end = neweprsd;
						growing++;
					}
					else /* matching stops, reset window pointers */
					{
						float percent = 0;
						//fprintf(STDOUT, "\nintertask window stops...growing = %d  of %d", growing, it_l_match_begin->getLoopSize());
#if 0
						/* if it is 80% match, then merge */
						percent = growing*100.0/(it_l_match_begin->getLoopSize());
						if(percent > PERCENT_MATCH)
						{
							//fprintf(STDOUT, "\nintertask EPRSD merging - partial match");
							intertask_mergeLeftRightPortions();							
							it_prev_merge_point = it_l_match_end;
						}
						else
						{
							/* include the new entry added as tail in the sequence */
							it_r_match_end = neweprsd;
							/* add sequence after in_prev_merge_point */
							intertask_addSequenceAfter(it_prev_merge_point);
							it_prev_merge_point = neweprsd;

						}
#endif
						it_l_match_begin = NULL;
						it_r_match_begin = NULL;
						it_l_match_end = NULL;
						it_r_match_end = NULL;
						growing = 0;
					}
					/* merge the EPRSD sequence */
					/* EPRSD sequence completed */
					if(growing)
					{
						if(it_l_match_begin->getLoopSize() == growing)
						{
							//fprintf(STDOUT, "\n intertask EPRSD merging");
							intertask_mergeLeftRightPortions();
							it_prev_merge_point = it_l_match_end;
							/* reset the pointers as they are no more required */
							it_l_match_begin = NULL;
							it_r_match_begin = NULL;
							it_l_match_end = NULL;
							it_r_match_end = NULL;
							growing = 0;
						}
					}
				}//else ends
#endif				
			}//else if ends
			else if(match->getAttr() & (ATTR_MEMBER|ATTR_TAIL)) /* in case growing stopped, put it next to its predecessors*/
			{
				if(it_l_match_begin || it_l_match_end || it_r_match_begin || it_r_match_end)
					fprintf(STDOUT, "\nERROR: IT: L & R not NULL#1");
				//addTail(neweprsd);
				addAfter(neweprsd, it_prev_merge_point);
				it_prev_merge_point = neweprsd;
			}
			else if(match->getAttr() == ATTR_NONE)/* immediately merge the RSDs*/
			{	
				if(it_l_match_begin || it_l_match_end || it_r_match_begin || it_r_match_end)
					fprintf(STDOUT, "\nERROR: IT: L & R not NULL#2");
				intertask_merge(match, neweprsd);
				it_prev_merge_point = match;
				delete neweprsd; /* neweprsd not needed after merging */
			}
			else
			{
				fprintf(STDOUT, "\nERROR: Invalid attribute.....");
			}
#endif
		}//if(match) ends
		else
		{	
#if 0
			if(neweprsd->getAttr() &(ATTR_LEADER|ATTR_MEMBER|ATTR_TAIL))
			{
				fprintf(STDOUT, "\nLoop leader/member/tail didn't find match");
			}
#endif
			//fprintf(STDOUT, "\nInter task - match NOT found for %p",(void*)neweprsd->cmninfo->ref_id);			
			//addTail(neweprsd);
			addAfter(neweprsd, it_prev_merge_point);
			it_prev_merge_point = neweprsd;
		}
		break;
	}//while ends
#endif

	//if an EPRSD, add as tail and look for the complete sequence match
	//when whole sequence is matched, merge the EPRSD
}

/* merge the multiple matching EPRSDs to one */
	template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::finalize()
{
	EPRSD<T> *match = NULL;
	EPRSD<T> *start = NULL;
	EPRSD<T> *neweprsd = NULL;
	EPRSD<T> *cur = NULL;
	EPRSD<T> *right_cur = NULL;
	EPRSD<T> *nxt = NULL;	
	EPRSD<T> *nxtsrch = NULL;	
	_INT64 stride = 0;
	_INT64 length = 0;
	_INT64 count = 0;
	_INT64 gap = 0;
	fprintf(STDOUT, "\n\n*****************finalize called*********************");
	fflush(STDOUT);

	/* takes care of left over nodes once trace stops e.g: xyz(4)abc(2)abc and trace stops. Following 5-6 lines of code
	   consolidates it to xyz(4)abc(3)*/
#if 0
	if(l_match_begin && l_match_end && r_match_begin && r_match_end)
	{	
		if(l_match_end->next == r_match_begin)
		{
			fprintf(STDOUT, "\nLEFTOVER nodes merged");
			mergeLeftRightPortions();
		}
	}
#endif
	l_match_begin = l_match_end = NULL;
	r_match_begin = r_match_end = NULL;	
	start = listhead;
	neweprsd = start;

	while(neweprsd)
	{
		//fprintf(STDOUT, "\nfinalize neweprsd: %p", neweprsd->cmninfo->ref_id);
		//handle RSD and EPRSDs differently
		//if(neweprsd->getCode() == CODE_RSD)
#if 0 //finalize hangs bug
		if(neweprsd->getAttr() & (ATTR_LEADER|ATTR_MEMBER|ATTR_TAIL))
		{
			match = findExactMatch(neweprsd, &gap);
			while(match)
			{
				//fprintf(STDOUT, "\nfinalize (RSD) src refid: %p, dest refid: %p", neweprsd->cmninfo->ref_id, match->cmninfo->ref_id);
				//fprintf(STDOUT, "ptrs: %p %p", neweprsd, match);
				//merge matched entries, assign eprsd being searched and matched into Left & Right portions respectively
				//update the length of RSDs
				cur = l_match_begin = neweprsd;
				right_cur = r_match_begin = match;
				count = l_match_begin->getLoopSize();
				if(l_match_begin && r_match_begin)
				{
					while(count--)
					{
						length = cur->getLength() + right_cur->getLength();
						cur->setLength(length);
						cur->setEprsdMergeCount(cur->getEprsdMergeCount()+1);
						if(neweprsd->getEprsdMergeCount()+1)
							neweprsd->gap = gap/(neweprsd->getEprsdMergeCount()+1);
						//printf("\nLOOP: gap = %ld, avg. gap = %ld, avg. length = %ld", gap, neweprsd->gap, (neweprsd->getLength() + match->getLength())/2 );
						l_match_end = cur;
						r_match_end = right_cur;
						cur = cur->next;
						right_cur = right_cur->next;
					}
				}//if ends

				//fprintf(STDOUT, "\n#finalize #(a)");
				if(r_match_begin && r_match_end)
				{
					if(r_match_begin->prev && r_match_end)	
						(r_match_begin->prev)->next = r_match_end->next;
					if(r_match_end->next && r_match_begin)
						(r_match_end->next)->prev = r_match_begin->prev;

					//fprintf(STDOUT, "\n#finalize #(b)");	
					if(listtail == r_match_end && r_match_begin)
						listtail = r_match_begin->prev;
					if(listhead == r_match_begin && r_match_end)
						listhead = r_match_end->next;

					r_match_begin->prev = NULL;	
					r_match_end->next = NULL;
				}

				//free the right portion
				cur = r_match_begin;
				while(cur && cur != r_match_end->next)
				{
					nxt = cur->next;
					//fprintf(STDOUT, "\nfinalize freeing %p, ", cur);	
					delete cur;
					//fprintf(STDOUT, "\nfinalize freed %p, ", cur);	
					cur = nxt;
				}
				//search again
				match = findExactMatch(neweprsd, &gap);
				nxtsrch = l_match_end->next;		 
				l_match_begin = l_match_end = NULL;	
				r_match_begin = r_match_end = NULL;	
			}//while(match) ends
			//fprintf(STDOUT, "\nJumping to the end of EPRSD");
			neweprsd = nxtsrch;		 	
			continue;
		}//if ATTR_LEADER|ATTR_MEMBER ends
		else
#endif
			if(neweprsd->getAttr() & (ATTR_NONE|ATTR_LEADER|ATTR_MEMBER|ATTR_TAIL) )
			{
				match = findExactMatch(neweprsd, &gap);
				while(match)
				{
					//fprintf(STDOUT, "\nfinalize (NODE) src refid: %p, dest refid: %p, avg. gap = %d", \
					neweprsd->cmninfo->ref_id, match->cmninfo->ref_id, neweprsd->gap);	 
					//fprintf(STDOUT, "ptrs: %p %p", neweprsd, match);
					//merge matched entries, assign eprsd being searched and matched into Left & Right portions respectively
					length = neweprsd->getLength() + match->getLength();
					neweprsd->setLength(length);
					neweprsd->setEprsdMergeCount(neweprsd->getEprsdMergeCount()+1);
					if(neweprsd->getEprsdMergeCount()+1)
						neweprsd->gap = gap/(neweprsd->getEprsdMergeCount()+1);
					//printf("\nRSD: gap = %ld, avg. gap = %ld, avg. length = %ld", gap, neweprsd->gap, (neweprsd->getLength() + match->getLength())/2 );
					//stride = (_abs( ((neweprsd->userinfo)->getValue()) - ((match->userinfo)->getValue()) ))/((neweprsd->getLength()) - 1);
					neweprsd->setStride(match->getStride());
					//remove the matched node
					if(listhead != match && match->prev)
						(match->prev)->next = match->next;
					else
						listhead = match->next;	

					if(listtail !=  match && match->next)
						(match->next)->prev = match->prev;
					else
						listtail = match->prev;

					delete match;
					match = findExactMatch(neweprsd, &gap);
				}
				neweprsd = neweprsd->next;
				continue;
			}//else if ATTR_NONE ends
			else
			{
				fprintf(STDOUT, "\nError: finalize - invalid attribute");
				break; //exit(-5);
			}	
	}//while ends
	return;
}

/* abstract class making getValue() and setValue() pure virtual functions - user must define them*/
class _INFO
{
	protected:
		_UINT64 value;

	public:
		_INFO()
		{
			value = 0;		
		}

		_INFO(_UINT64 addr)
		{
			//cout<<"\nINFO ctor";
			value = addr;
		}

		~_INFO()
		{
			//cout<<"\nINFO dtor";
		}

		virtual _UINT64 getValue() = 0;
		virtual void setValue(void* val) = 0;
		//must overload - operator
};



