/*
 *  Doubly Linked List template class
 *  Author: Sandeep Budanur Ramanna
 *  E-mail: sbudanu@ncsu.edu
 *  North Carolina State University
 *  Created on Dec 01 2009
 */
//varibales: separate words by underscore
//functions: first word in smaller case, only first letter of subsequent words in uppercase
#ifndef __MYQUEUE_CPP
#define __MYQUEUE_CPP
#include <iostream>
using namespace std;
typedef long _INT64;

template <class U>
class LIST_ITERATOR;

template <class U>
class LIST_RITERATOR;

template <class M>
struct NODE
{
	M data;
	struct NODE *prev;
	struct NODE *next;
};	

template <class T>
class LIST
{
	private:

		struct NODE<T>* head;
		struct NODE<T>* tail;
		struct NODE<T>* iterator;
		struct NODE<T>* riterator;
		_INT64 count;
		_INT64 createIterCount;
		_INT64 createRIterCount;
		_INT64 releaseIterCount;
		_INT64 releaseRIterCount;

	public:
		template<class U>friend class LIST_ITERATOR;
		template<class U>friend class LIST_RITERATOR;
		LIST();
		~LIST();
		bool addTail(const T& data);
		bool addHead(const T& data);
		LIST_ITERATOR<T>* createIterator();
		LIST_RITERATOR<T>* createRIterator();
		bool insert(LIST_ITERATOR<T>* pos, const T& data);
		void removeHead(T& data);
		void removeTail(T& data);
		void removeTail();
		void removeHead();
		void removeAll();
		void getTail(T& out);
		_INT64 getSize();
		void releaseIterator(LIST_ITERATOR<T>* iter);
		void releaseRIterator(LIST_RITERATOR<T>* riter);
		void printIteratorStatistics();
		//bool setIteratorNext();
		//bool setRIteratorNext();
		//bool getIteratedItem(T& item);
};

template <class U>
class LIST_ITERATOR
{
	private:
		const LIST<U>* list;
		struct NODE<U>* iter;

	public:
		LIST_ITERATOR(const LIST<U> *l)
		{
			list = l;
			iter = NULL;
		}

		~LIST_ITERATOR()
		{
			list = NULL;
			iter = NULL;
		}	

		bool next()
		{
			if(iter == list->tail)
			{
				iter = NULL;
				return false;
			}
			else if(!iter)
			{
				iter = list->head;
				return true;
			}
			else
			{
				iter = iter->next;
				return true;
			}
			return false;
		}

		U& currentItem()
		{
			return iter->data;
		}

		struct NODE<U>* getCurrentPos()
		{
			return iter;
		}

};

template <class U>
class LIST_RITERATOR
{
	private:
		const LIST<U>* list;
		struct NODE<U>* iter;

	public:
		LIST_RITERATOR(const LIST<U> *l)
		{
			list = l;
			iter = NULL;
		}

		bool next()
		{
			if(iter == list->head)
			{
				iter = NULL;
				return false;
			}
			else if(!iter)
			{
				iter = list->tail;
				return true;
			}
			else
			{
				iter = iter->prev;
				return true;
			}
			return false;
		}

		U& currentItem()
		{
			return iter->data;
		}

		struct NODE<U>* getCurrentPos()
		{
			return iter;
		}
};


	template <class T>
LIST<T>::LIST()
{
	//cout<<"List Constructor\n";	
	head = tail = NULL;
	iterator = NULL;
	count = 0;
}


	template <class T>
LIST<T>::~LIST()
{
	struct NODE<T>* cur = head;
	struct NODE<T>* nxt = head;
	//printf("\ndtor");
	while(cur)
	{
		nxt = cur->next;
		delete cur;
		cur = nxt;
	}
	head = tail = NULL;
	count = 0;
	//printIteratorStatistics();
}

/* needed by memtrace */
	template <class T>
void LIST<T>::removeHead(T& data)
{
	if(head == NULL || tail == NULL)
		return;
	else if(head == tail)/* one node */
	{
		data = head->data;
		delete head;
		head = tail = NULL;
		count--;
	}
	else
	{
		data = head->data;
		head = head->next;
		delete head->prev;
		head->prev = NULL;
		count--;
	}
	return;		
}

/* needed by memtrace */
	template <class T>
void LIST<T>::removeTail(T& data)
{
	if(head == NULL || tail == NULL)
		return;
	else if(head == tail)/* one node */
	{
		data = tail->data;
		delete tail;
		head = tail = NULL;
		count--;
	}
	else
	{
		data = tail->data;
		tail = tail->prev;
		delete tail->next;
		tail->next = NULL;
		count--;
	}
	return;	
}

	template <class T>
void LIST<T>::removeHead()
{
	struct NODE<T>* cur = head;

	if(!cur) /* empty list */
		return;	

	head = cur->next;

	if(head)
		head->prev = NULL;
	else /* list had only one node */
		tail = NULL;

	count--;
	delete cur;
}


	template <class T>
void LIST<T>::removeTail()
{
	struct NODE<T>* cur = tail;

	if(!cur) /* empty list */
		return;	

	tail = cur->prev;

	if(tail)
		tail->next = NULL;
	else /* list had only one node */
		head = NULL;

	count--;
	delete cur;
}


	template <class T>
void LIST<T>::removeAll()
{
	struct NODE<T>* cur = head;
	struct NODE<T>* nxt = NULL;
	//printf("\nremove");
	while(cur)
	{
		nxt = cur->next;
		delete cur;
		cur = nxt;
	}
	head = tail = NULL;
	count = 0;	
}

	template <class T>
_INT64 LIST<T>::getSize()
{
	return count;
}

	template <class T>
void LIST<T>::getTail(T& out)
{
	if(tail)	
		out = tail->data;
	return;	
}

	template <class T>
LIST_ITERATOR<T>* LIST<T>::createIterator()
{
	createIterCount++;
	return new LIST_ITERATOR<T>(this);
}

	template <class T>
LIST_RITERATOR<T>* LIST<T>::createRIterator()
{
	createRIterCount++;
	return new LIST_RITERATOR<T>(this);
}

	template <class T>
void LIST<T>::releaseIterator(LIST_ITERATOR<T>* iter)
{
	releaseIterCount++;
	delete iter;
}

	template <class T>
void LIST<T>::releaseRIterator(LIST_RITERATOR<T>* riter)
{
	releaseRIterCount++;
	delete riter;
}


	template <class T>
bool LIST<T>::addTail(const T& data)
{
	NODE<T> *item  = new(NODE<T>);
	item->data = data;
	item->prev = NULL;
	item->next = NULL;

	if(!head || !tail)//list is empty
	{
		head = item;
		tail = item;		
	}
	else //add to the end
	{
		item->prev = tail;
		tail->next = item;		
		tail = item;
	}	 
	count++;
	return true;	
}


	template <class T>
bool LIST<T>::addHead(const T& data)
{
	NODE<T> *item  = new(NODE<T>);
	item->data = data;
	item->prev = NULL;
	item->next = NULL;

	if(!head || !tail)//list is empty
	{
		head = item;
		tail = item;
	}
	else //add to the end
	{
		item->next = head;
		head->prev = item;
		head = item;
	}
	count++;
	return true;
}

//insert data before pos
	template <class T>
bool LIST<T>::insert(LIST_ITERATOR<T>* pos, const T& data)
{
	if(pos->getCurrentPos() == NULL)
		return false;
	if(pos->getCurrentPos() == head)
		return addHead(data);
	else
	{
		struct NODE<T>* cur = pos->getCurrentPos();
		NODE<T> *item  = new(NODE<T>);
		item->data = data;
		item->prev = NULL;
		item->next = NULL;		
		item->next = cur;
		item->prev = cur->prev;
		(cur->prev)->next = item;
		cur->prev = item;
		count++;
		return true;
	}

}

	template <class T>
void LIST<T>::printIteratorStatistics()
{
	printf("\nLIST ITERATOR STATISTICS");
	printf("\n========================================");
	printf("\nITER>>>");
	printf("\nCreated: %d times", createIterCount);
	printf("\nReleased: %d times", releaseIterCount);
	printf("\nMemory leak: %d", createIterCount - releaseIterCount);
	printf("\nRITER>>>");
	printf("\nCreated: %d times", createRIterCount);
	printf("\nReleased: %d times", releaseRIterCount);
	printf("\nMemory leak: %d", createRIterCount - releaseRIterCount);
}

#if 0
	template <class T>
bool LIST<T>::setIteratorNext()
{
	if(iterator == tail)
	{
		iterator = NULL; //reset on reaching end
		cout<<"ireset";
		return false;
	}
	else if(!iterator)
	{
		iterator = head;
		return true;
	}
	else
	{
		iterator  = iterator->next;
		return true;
	}
}

	template <class T>
bool LIST<T>::setRIteratorNext()
{
	if(riterator == head)
	{
		iterator = NULL; //reset on reaching end
		cout<<"rireset";
		return false;
	}
	else if(!riterator)
	{
		riterator = tail;
		return true;
	}
	else
	{
		iterator  = iterator->prev;
		return true;
	}

}

	template <class T>
bool LIST<T>::getIteratedItem(T& item)
{
	if(iterator)
	{
		item = iterator->data;
		return true;
	}

	return false;
}
#endif

#if 0
int main()
{
	LIST<int> l;
	LIST_ITERATOR<int> *iter = l.createIterator();
	LIST_RITERATOR<int> *riter = l.createRIterator();
	LIST_ITERATOR<int> *iterA = l.createIterator();

	int n=100;
	int i=0;
	l.addTail(100);
	l.addHead(99);
	l.addHead(98);
	l.addTail(101);

	iterA->next();
	iterA->next();
	iterA->next();
	iterA->next();
	//iterA->next();
	l.insert(iterA,8888);	

	while(iter->next())
	{
		cout<<"*"<<iter->currentItem()<<"*";		
	}

	cout<<endl;

	while(riter->next())
	{
		cout<<"*"<<riter->currentItem()<<"*";		
	}

	cout<<endl;

#if 0
	LIST<double> l1;
	LIST_ITERATOR<double> *iter1 = l1.createIterator();
	LIST_RITERATOR<double> *riter1 = l1.createRIterator();

	double n1=12.14;
	double i1=0.0;
	l1.addTail(14.657);
	l1.addHead(213.87667);
	l1.addHead(78.4534);
	l1.addTail(6756.65546);

	while(iter1->next())
	{
		cout<<"*"<<iter1->currentItem()<<"*";
	}

	cout<<endl;

	while(riter1->next())
	{
		cout<<"*"<<riter1->currentItem()<<"*";
	}

	cout<<endl;
#endif


#if 0
	while(l.setIteratorNext())
	{
		l.getIteratedItem(i);
		cout<<"*"<<i<<"*";
	}
	cout<<endl;	
	while(l.setRIteratorNext())
	{
		l.getIteratedItem(i);
		cout<<"*"<<i<<"*";
	}
	cout<<endl;
#endif
	return 0;
}
#endif

#endif
