/* 
 * Tool to replay compressed EPRSDs 
 * Author: Sandeep Budanur Ramanna, sbudanu@ncsu.edu
 * North Carolina State University
 * Created on 14th July 2010
 * 
 * */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

/* for structure declarations and definitions */
#include "myprsdtlib.h"
//#include "types.h"

//#define LOOPHEAD

typedef struct
{
	_INT64 attr;
	_INT64 gap;
	_UINT64 start_value;
	struct details det[3];
	_UINT64 sign_len;
	_UINT64 signature[MAX_SIGN_LEN];
	_UINT64 ref_id;
	_UINT64 type;
}EPRSD_INFO;

typedef struct RRNODE RRNODE;

struct RNODE
{
	EPRSD_INFO *data;
	RNODE *prev;
	RNODE *next;
};

class RLIST
{
	public:
		RNODE *head;
		RNODE *tail;
		_INT64 count;

	public:
		RLIST()
		{
			head = tail = NULL;
			count = 0;
		}

		~RLIST()
		{
			removeAll();
		}

		void addTail(EPRSD_INFO *data)
		{
			RNODE *item  = new(RNODE);
			item->data = data;
			item->prev = NULL;
			item->next = NULL;

			if(!head || !tail)//list is empty
			{
				head = item;
				tail = item;		
			}
			else //add to the end
			{
				item->prev = tail;
				tail->next = item;
				tail = item;
			}
			count++;
			return;	
		}

		void removeAll()
		{
			RNODE *cur = NULL;
			RNODE *nxt = NULL;
			cur = head;
			while(cur)
			{
				nxt = cur->next;
				free(cur->data); /* malloced while reading trace from file */
				delete cur;
				cur = nxt;
				count--;
			}
			head = tail = NULL;
			return;
		}

		void display()
		{
			RNODE *cur = NULL;
			fprintf(stdout, "\nStack Contents from bottom");
			cur = head;

			while(cur)
			{
				fprintf(stdout, "\n%d %p", cur->data->det[0].length, cur->data->det[0].start_value);
				cur = cur->next;
			}
			fprintf(stdout, "\n");
		}
};

EPRSD_INFO* readEntry();

char filename[256];
FILE* fptr = NULL;
RLIST list;
//EPRSD_INFO *info;

/* read an EPRSD entry from input file and returns a pointer to dynamically allocated
 * memory containing EPRSD details. Caller must free the memory allocated by invoking free().
 */
EPRSD_INFO* readEntry()
{
	EPRSD_INFO *myinfo = NULL;
	EPRSD_INFO info;
	char signstr[10];
	int ret = 0, i = 0;
	myinfo = (EPRSD_INFO*)malloc(sizeof(EPRSD_INFO));
	if(!myinfo)
	{
		fprintf(stderr, "\nmalloc error - info");
		exit(-3);
	}

	do
	{
#if 0
		fprintf(outfptr, "\n\n%ld %ld %ld", attr,  det[0].loopsize, det[0].length);
		fprintf(outfptr, "\n%ld", gap);
		fprintf(outfptr, "\n%p", (void*)det[0].start_value);
		fprintf(outfptr, "\n%ld", det[0].length);
		fprintf(outfptr, "\n%p", (void*)det[0].stride);
		fprintf(outfptr, "\n%p", (void*)det[0].tid_start_value);
		fprintf(outfptr, "\n%ld", det[0].tid_length);
		fprintf(outfptr, "\n%p", (void*)det[0].tid_stride);
		fprintf(outfptr, "\n%p", (void*)det[0].tid_bit_pattern);
		fprintf(outfptr, "\n%p", (void*)det[0].tid_addr_stride);
		fprintf(outfptr, "\n%p", (void*)det[0].node_id_start_value);
		fprintf(outfptr, "\n%ld", det[0].node_id_length);
		fprintf(outfptr, "\n%p", (void*)det[0].node_id_stride);
		fprintf(outfptr, "\n%p", (void*)det[0].nodeid_bit_pattern);
#else
		ret = fscanf(fptr, "\n\nattr: %ld loopsize: %ld loopcount: %ld, sign: %p", &info.attr,  &info.det[0].loopsize, &info.det[0].loopcount, &info.ref_id);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\ngap: %ld", &info.gap);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\nstart_value: %p", &info.det[0].start_value);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\nlength: %ld", &info.det[0].length);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\nstride: %p", &info.det[0].stride);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\ntid_start_value: %p", &info.det[0].tid_start_value);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\ntid_length: %ld", &info.det[0].tid_length);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\ntid_stride: %p", &info.det[0].tid_stride);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\ntid_bit_pattern: %p", &info.det[0].tid_bit_pattern);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\ntid_addr_stride: %p", &info.det[0].tid_addr_stride);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\nnode_id_start_value: %p", &info.det[0].node_id_start_value);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\nnode_id_length: %ld", &info.det[0].node_id_length);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\nnode_id_stride: %p", &info.det[0].node_id_stride);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		ret = fscanf(fptr, "\nnode_id_bit_pattern: %p", &info.det[0].nodeid_bit_pattern);
		if(ret == 0 || ret == EOF){assert(ret);break;}	
		ret = fscanf(fptr, "\nSign_len = %ld", &info.sign_len);
		if(ret == 0 || ret == EOF){assert(ret);break;}	
		ret = fscanf(fptr, "\nSign:%p ", &info.signature[0]);
		if(ret == 0 || ret == EOF){assert(ret);break;}
		//printf("\n^^^^%s^^^^", signstr);
		//printf("\n****%d****", info.sign_len);	
		//printf("\n####%p####", info.det[0].start_value);	
		//printf("\n^^^^%p^^^^", info.signature[0]);	
		for(i=1; i<info.sign_len; i++)
		{
			ret = fscanf(fptr, "%p ", &info.signature[i]);		
			//printf("\n$$$$%p$$$$", info.signature[i]);
			if(ret == 0 || ret == EOF){assert(ret);break;}	
		}
		if(ret == 0 || ret == EOF){assert(ret);break;}	
		ret = fscanf(fptr, "\nref_id = %p",&info.ref_id);		
		if(ret == 0 || ret == EOF){assert(ret);break;}	
		ret = fscanf(fptr, "\ntype = %p", &info.type);
		if(ret == 0 || ret == EOF){assert(ret);break;}	
		//getchar();
#endif
#if 0		
		fprintf(stdout, "\n\n%ld %ld %ld", info.attr,  info.det[0].loopsize, info.det[0].length);
		fprintf(stdout, "\n%ld", info.gap);
		fprintf(stdout, "\n%p", (void*)info.det[0].start_value);
		fprintf(stdout, "\n%ld", info.det[0].length);
		fprintf(stdout, "\n%p", (void*)info.det[0].stride);
		fprintf(stdout, "\n%p", (void*)info.det[0].tid_start_value);
		fprintf(stdout, "\n%ld", info.det[0].tid_length);
		fprintf(stdout, "\n%p", (void*)info.det[0].tid_stride);
		fprintf(stdout, "\n%p", (void*)info.det[0].tid_bit_pattern);
		fprintf(stdout, "\n%p", (void*)info.det[0].tid_addr_stride);
		fprintf(stdout, "\n%p", (void*)info.det[0].node_id_start_value);
		fprintf(stdout, "\n%ld", info.det[0].node_id_length);
		fprintf(stdout, "\n%p", (void*)info.det[0].node_id_stride);
		fprintf(stdout, "\n%p", (void*)info.det[0].nodeid_bit_pattern);	
		fprintf(stdout, "\n%ld", info.sign_len);
		fprintf(stdout, "\nSign:");

		for(i=0; i<info.sign_len; i++)
			fprintf(stdout, "%p ", (void*)info.signature[i]);

		fprintf(stdout, "\n%p",(void*)info.ref_id);
		fprintf(stdout, "\n%p", (void*)info.type);
		getchar();
#endif
	}while(0);

	if(ret == 0 || ret == EOF)
		return NULL;
	else
	{
		*myinfo = info;
		return myinfo;
	}
}

RNODE* print(RNODE* cur)
{
	_INT64 node_id;
	_INT64 tid;
	_INT64 addr;
	_INT64 s, a, t, n;
	EPRSD_INFO* info = NULL;

	assert(cur);
	info = cur->data;
	assert(info);

	for(n = 0, node_id = info->det[0].node_id_start_value; n < info->det[0].node_id_length; n++, node_id += info->det[0].node_id_stride)
	{
		for(t = 0, tid = info->det[0].tid_start_value; t < info->det[0].tid_length; t++, tid += info->det[0].tid_stride)
		{
			//fprintf(stdout, "\n***%d***", info->det[0].length);
			for(a = 0, addr = info->det[0].start_value; a < info->det[0].length; a++, addr += info->det[0].stride)
			{
				fprintf(stdout, "\n\nnode id: %d", node_id);
				fprintf(stdout, "\ntid: %d", tid);
				fprintf(stdout,"\naddr:%p type:%c", addr, info->type);
				fprintf(stdout,"\nSign_len: %d", info->sign_len);
				fprintf(stdout,"\nSign:");
				for(s = 0; s < info->sign_len; s++)
				{
					fprintf(stdout, "%p ", info->signature[s]);
				}
			}
		}
	}
	return cur;
}

RNODE* printSingle(RNODE* cur, _INT64 iter)
{
	_INT64 node_id;
	_INT64 tid;
	_INT64 addr;
	_INT64 s, a, t, n;
	EPRSD_INFO* info = NULL;

	assert(cur);
	info = cur->data;
	assert(info);

	for(n = 0, node_id = info->det[0].node_id_start_value; n < info->det[0].node_id_length; n++, node_id += info->det[0].node_id_stride)
	{
		for(t = 0, tid = info->det[0].tid_start_value; t < info->det[0].tid_length; t++, tid += info->det[0].tid_stride)
		{
			//fprintf(stdout, "\n***%d***", info->det[0].length);
			//for(a = 0, addr = info->det[0].start_value; a < info->det[0].length; a++, addr += info->det[0].stride)
			//{
			addr = info->det[0].start_value + iter * info->det[0].stride;
			fprintf(stdout, "\n\nnode id: %d", node_id);
			fprintf(stdout, "\ntid: %d", tid);
			fprintf(stdout,"\naddr:%p type:%c", addr, info->type);
			fprintf(stdout,"\nSign_len: %d", info->sign_len);
			fprintf(stdout,"\nSign:");
			for(s = 0; s < info->sign_len; s++)
			{
				fprintf(stdout, "%p ", info->signature[s]);
			}
			//}
		}
	}
	return cur;

}

RNODE* printLoop(RNODE* loophead, _INT64* items)
{
	RNODE* cur = NULL;
	RNODE* prev = NULL;
	_INT64 i = 0, j = 0;
	bool tail_reached  = 0;

	//fprintf(stdout, "**********Entering print loop************\n");	
	for(i = 0; i < loophead->data->det[0].loopcount; i++)
	{
		//fprintf(stdout, "\n************Printing head - %d***************", i);
		cur = loophead->next;
		cur = printSingle(cur, i);
		prev = cur;
		cur = cur->next;
		for(j = 1; j < loophead->data->det[0].loopsize-1; j++)
			//while(1)
		{
			if(cur->data->attr & ATTR_LEADER)
			{
				//fprintf(stdout, "\nj before printLoop = %d", j);
#ifdef LOOPHEAD
				cur = cur->next;
#endif	
				cur = printLoop(cur, &j);
				j--; /* it is incremented w/o printing, adjust */
				//fprintf(stdout, "\nj after printLoop = %d", j);	
			}
			else
			{
				cur = printSingle(cur, i);
				//fprintf(stdout, "\nj after printSingle = %d", j);	
#if 0
				if(cur->data->attr & ATTR_TAIL) /* loop end */
				{
					tail_reached = 1;
					//fprintf(stdout, "\n**********Tail reached***************");
					//break;		
				}
#endif
			}
			prev = cur;
			cur = cur->next;			
		}
	}

	(*items) += loophead->data->det[0].loopsize;

	if(cur == loophead)
	{
		fprintf(stdout, "\nError: list traversed incorrectly");
		return cur;
	}
	/*
	   else if(!tail_reached)
	   {
	//fprintf(stdout, "\nError: returning before loop tail is reached");
	return cur->prev;
	}
	*/
	else /* would have reached one node ahead after finishing loop */
	{
		//fprintf(stdout, "**********Exiting print loop************\n");	
		return prev;
	}
}

int main(int argc, char** argv)
{
	int ret = 0;
	int count = 0;
	_INT64 items = 0;

	EPRSD_INFO *info = NULL;
	RNODE *cur = NULL;

	if(argc != 2)
	{
		fprintf(stderr, "\nUsage:replay <eprsd_file>\n");
		exit(-1);
	}

	strcpy(filename, argv[1]);
	//printf("\nfile: %s\n", filename);
	if((fptr = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "\ninput file open error: %s", filename);
		exit(-2);
	}

	do
	{
		info = readEntry();
		if(!info)
			break;
		count++;
		list.addTail(info);
		//replayTrace(info);
		//getchar();
	}while(1);

	fprintf(stdout, "\n# of EPRSDs read = %d", count);

	cur = list.head;
	while(cur)
	{
		items = 0;
		if(cur->data->attr == ATTR_NONE)
		{
			cur = print(cur);			
		}
		else if(cur->data->attr & (ATTR_LEADER|ATTR_NONE))
		{
#ifdef LOOPHEAD
			cur = cur->next;
#endif
			cur = printLoop(cur, &items);//returns the pointer to the last node in the loop
			//fprintf(STDOUT, "\n$$$$$%d items in loop$$$$$", items);
			if(!cur)
			{
				fprintf(stderr, "\nError: printLoop() returned NULL");
				exit(-2);
			}
		}
		else
		{
			fprintf(stderr, "\nError: MEMBER found without LEADER");
			exit(-1);
		}
		cur = cur->next;
	}
#if 0
	if(ret == EOF)
	{
		fprintf(stdout, "\nEOF reached");
	}
	else if(ret == 0)
	{
		fprintf(stdout, "\nfscanf error");
	}
#endif
	return 0;
}
